import duckdb
import os
import datetime
import pytest
import pandas as pd
from conftest import pandas_2_or_higher


class TestPandasTimestamps(object):
    @pytest.mark.parametrize('unit', ['s', 'ms', 'us', 'ns'])
    def test_timestamp_types_roundtrip(self, unit):
        d = {
            'time': pd.Series(
                [pd.Timestamp(datetime.datetime(2020, 6, 12, 14, 43, 24, 394587), unit=unit)],
                dtype=f'datetime64[{unit}]',
            )
        }
        df = pd.DataFrame(data=d)
        df_from_duck = duckdb.from_df(df).df()
        assert df_from_duck.equals(df)

    @pytest.mark.parametrize('unit', ['s', 'ms', 'us', 'ns'])
    def test_timestamp_timezone_roundtrip(self, unit):
        if pandas_2_or_higher():
            dtype = pd.core.dtypes.dtypes.DatetimeTZDtype(unit=unit, tz='UTC')
            expected_dtype = pd.core.dtypes.dtypes.DatetimeTZDtype(unit='us', tz='UTC')
        else:
            # Older versions of pandas only support 'ns' as timezone unit
            expected_dtype = pd.core.dtypes.dtypes.DatetimeTZDtype(unit='ns', tz='UTC')
            dtype = pd.core.dtypes.dtypes.DatetimeTZDtype(unit='ns', tz='UTC')

        conn = duckdb.connect()
        conn.execute("SET TimeZone =UTC")
        d = {
            'time': pd.Series(
                [pd.Timestamp(datetime.datetime(2020, 6, 12, 14, 43, 24, 394587), unit=unit, tz='UTC')],
                dtype=dtype,
            )
        }
        df = pd.DataFrame(data=d)

        # Our timezone aware type is in US (microseconds), when we scan a timestamp column that isn't US and has timezone info,
        # we convert the time unit to US
        expected = pd.DataFrame(data=d, dtype=expected_dtype)
        df_from_duck = conn.from_df(df).df()
        assert df_from_duck.equals(expected)

    @pytest.mark.parametrize('unit', ['s', 'ms', 'us', 'ns'])
    def test_timestamp_nulls(self, unit):
        d = {'time': pd.Series([pd.Timestamp(None, unit=unit)], dtype=f'datetime64[{unit}]')}
        df = pd.DataFrame(data=d)
        df_from_duck = duckdb.from_df(df).df()
        assert df_from_duck.equals(df)

    def test_timestamp_timedelta(self):
        df = pd.DataFrame(
            {
                'a': [pd.Timedelta(1, unit='s')],
                'b': [pd.Timedelta(None, unit='s')],
                'c': [pd.Timedelta(1, unit='us')],
                'd': [pd.Timedelta(1, unit='ms')],
            }
        )
        df_from_duck = duckdb.from_df(df).df()
        assert df_from_duck.equals(df)

    def test_timestamp_timezone(self, duckdb_cursor):
        rel = duckdb_cursor.query(
            """
            SELECT
                '2019-01-01 00:00:00+00'::TIMESTAMPTZ AS dateTime,
                '2019-01-01 00:00:00+00'::TIMESTAMPTZ AS dateTime
        """
        )
        res = rel.df()
        assert res['dateTime'][0] == res['dateTime_1'][0]



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")
from duckdb.experimental.spark.sql.types import Row
from duckdb.experimental.spark.sql.types import (
    StringType,
    BinaryType,
    BitstringType,
    UUIDType,
    BooleanType,
    DateType,
    TimestampType,
    TimestampNTZType,
    TimeType,
    TimeNTZType,
    TimestampNanosecondNTZType,
    TimestampMilisecondNTZType,
    TimestampSecondNTZType,
    DecimalType,
    DoubleType,
    FloatType,
    ByteType,
    UnsignedByteType,
    ShortType,
    UnsignedShortType,
    IntegerType,
    UnsignedIntegerType,
    LongType,
    UnsignedLongType,
    HugeIntegerType,
    UnsignedHugeIntegerType,
    DayTimeIntervalType,
    ArrayType,
    MapType,
    StructField,
    StructType,
)


class TestTypes(object):
    def test_all_types_schema(self, spark):
        # Create DataFrame
        df = spark.sql(
            """
			select * EXCLUDE (
				small_enum,
				medium_enum,
				large_enum,
				'union',
				fixed_int_array, 
				fixed_varchar_array, 
				fixed_nested_int_array,
            	fixed_nested_varchar_array, 
            	fixed_struct_array, 
            	struct_of_fixed_array, 
            	fixed_array_of_int_list,
                list_of_fixed_int_array
			) from test_all_types()
		"""
        )
        schema = df.schema
        assert schema == StructType(
            [
                StructField('bool', BooleanType(), True),
                StructField('tinyint', ByteType(), True),
                StructField('smallint', ShortType(), True),
                StructField('int', IntegerType(), True),
                StructField('bigint', LongType(), True),
                StructField('hugeint', HugeIntegerType(), True),
                StructField('uhugeint', UnsignedHugeIntegerType(), True),
                StructField('utinyint', UnsignedByteType(), True),
                StructField('usmallint', UnsignedShortType(), True),
                StructField('uint', UnsignedIntegerType(), True),
                StructField('ubigint', UnsignedLongType(), True),
                StructField('date', DateType(), True),
                StructField('time', TimeNTZType(), True),
                StructField('timestamp', TimestampNTZType(), True),
                StructField('timestamp_s', TimestampSecondNTZType(), True),
                StructField('timestamp_ms', TimestampNanosecondNTZType(), True),
                StructField('timestamp_ns', TimestampMilisecondNTZType(), True),
                StructField('time_tz', TimeType(), True),
                StructField('timestamp_tz', TimestampType(), True),
                StructField('float', FloatType(), True),
                StructField('double', DoubleType(), True),
                StructField('dec_4_1', DecimalType(4, 1), True),
                StructField('dec_9_4', DecimalType(9, 4), True),
                StructField('dec_18_6', DecimalType(18, 6), True),
                StructField('dec38_10', DecimalType(38, 10), True),
                StructField('uuid', UUIDType(), True),
                StructField('interval', DayTimeIntervalType(0, 3), True),
                StructField('varchar', StringType(), True),
                StructField('blob', BinaryType(), True),
                StructField('bit', BitstringType(), True),
                StructField('int_array', ArrayType(IntegerType(), True), True),
                StructField('double_array', ArrayType(DoubleType(), True), True),
                StructField('date_array', ArrayType(DateType(), True), True),
                StructField('timestamp_array', ArrayType(TimestampNTZType(), True), True),
                StructField('timestamptz_array', ArrayType(TimestampType(), True), True),
                StructField('varchar_array', ArrayType(StringType(), True), True),
                StructField('nested_int_array', ArrayType(ArrayType(IntegerType(), True), True), True),
                StructField(
                    'struct',
                    StructType([StructField('a', IntegerType(), True), StructField('b', StringType(), True)]),
                    True,
                ),
                StructField(
                    'struct_of_arrays',
                    StructType(
                        [
                            StructField('a', ArrayType(IntegerType(), True), True),
                            StructField('b', ArrayType(StringType(), True), True),
                        ]
                    ),
                    True,
                ),
                StructField(
                    'array_of_structs',
                    ArrayType(
                        StructType([StructField('a', IntegerType(), True), StructField('b', StringType(), True)]), True
                    ),
                    True,
                ),
                StructField('map', MapType(StringType(), StringType(), True), True),
            ]
        )



import duckdb
import pytest

# https://pandas.pydata.org/docs/dev/user_guide/copy_on_write.html
pandas = pytest.importorskip('pandas', '1.5', reason='copy_on_write does not exist in earlier versions')
import datetime


# Make sure the variable get's properly reset even in case of error
@pytest.fixture(autouse=True)
def scoped_copy_on_write_setting():
    old_value = pandas.options.mode.copy_on_write
    pandas.options.mode.copy_on_write = True
    yield
    # Reset it at the end of the function
    pandas.options.mode.copy_on_write = old_value
    return


def convert_to_result(col):
    return [(x,) for x in col]


class TestCopyOnWrite(object):
    @pytest.mark.parametrize(
        'col',
        [
            ['a', 'b', 'this is a long string'],
            [1.2334, None, 234.12],
            [123234, -213123, 2324234],
            [datetime.date(1990, 12, 7), None, datetime.date(1940, 1, 13)],
            [datetime.datetime(2012, 6, 21, 13, 23, 45, 328), None],
        ],
    )
    def test_copy_on_write(self, col):
        assert pandas.options.mode.copy_on_write == True
        con = duckdb.connect()
        df_in = pandas.DataFrame(
            {
                'numbers': col,
            }
        )
        rel = con.sql('select * from df_in')
        res = rel.fetchall()
        print(res)
        expected = convert_to_result(col)
        assert res == expected



import duckdb
import pytest
from duckdb import ColumnExpression


@pytest.fixture
def con():
    conn = duckdb.connect()
    # Main relation
    conn.execute(
        """
        create table tbl_a as (SELECT * FROM (VALUES
            (1, 1),
            (2, 1),
            (3, 2)
        ) AS t(a, b))
    """
    )

    # Other relation
    conn.execute(
        """
        create table tbl_b as (SELECT * FROM (VALUES
            (1, 4),
            (3, 5),
        ) AS t(a, b))
    """
    )
    yield conn


class TestRAPIJoins(object):
    def test_outer_join(self, con):
        a = con.table('tbl_a')
        b = con.table('tbl_b')
        expr = ColumnExpression('tbl_a.b') == ColumnExpression('tbl_b.a')
        rel = a.join(b, expr, 'outer')
        res = rel.fetchall()
        assert res == [(1, 1, 1, 4), (2, 1, 1, 4), (3, 2, None, None), (None, None, 3, 5)]

    def test_inner_join(self, con):
        a = con.table('tbl_a')
        b = con.table('tbl_b')
        expr = ColumnExpression('tbl_a.b') == ColumnExpression('tbl_b.a')
        rel = a.join(b, expr, 'inner')
        res = rel.fetchall()
        assert res == [(1, 1, 1, 4), (2, 1, 1, 4)]

    def test_anti_join(self, con):
        a = con.table('tbl_a')
        b = con.table('tbl_b')
        expr = ColumnExpression('tbl_a.b') == ColumnExpression('tbl_b.a')
        rel = a.join(b, expr, 'anti')
        res = rel.fetchall()
        # Only output the row(s) from A where the condition is false
        assert res == [(3, 2)]

    def test_left_join(self, con):
        a = con.table('tbl_a')
        b = con.table('tbl_b')
        expr = ColumnExpression('tbl_a.b') == ColumnExpression('tbl_b.a')
        rel = a.join(b, expr, 'left')
        res = rel.fetchall()
        assert res == [(1, 1, 1, 4), (2, 1, 1, 4), (3, 2, None, None)]

    def test_right_join(self, con):
        a = con.table('tbl_a')
        b = con.table('tbl_b')
        expr = ColumnExpression('tbl_a.b') == ColumnExpression('tbl_b.a')
        rel = a.join(b, expr, 'right')
        res = rel.fetchall()
        assert res == [(2, 1, 1, 4), (1, 1, 1, 4), (None, None, 3, 5)]

    def test_semi_join(self, con):
        a = con.table('tbl_a')
        b = con.table('tbl_b')
        expr = ColumnExpression('tbl_a.b') == ColumnExpression('tbl_b.a')
        rel = a.join(b, expr, 'semi')
        res = rel.fetchall()
        assert res == [(1, 1), (2, 1)]



import duckdb
import pytest
from conftest import NumpyPandas, ArrowPandas

closed = lambda: pytest.raises(duckdb.ConnectionException, match='Connection has already been closed')
no_result_set = lambda: pytest.raises(duckdb.InvalidInputException, match='No open result set')


class TestRuntimeError(object):
    def test_fetch_error(self):
        con = duckdb.connect()
        con.execute("create table tbl as select 'hello' i")
        with pytest.raises(duckdb.ConversionException):
            con.execute("select i::int from tbl").fetchall()

    def test_df_error(self):
        con = duckdb.connect()
        con.execute("create table tbl as select 'hello' i")
        with pytest.raises(duckdb.ConversionException):
            con.execute("select i::int from tbl").df()

    def test_arrow_error(self):
        pytest.importorskip('pyarrow')

        con = duckdb.connect()
        con.execute("create table tbl as select 'hello' i")
        with pytest.raises(duckdb.ConversionException):
            con.execute("select i::int from tbl").arrow()

    def test_register_error(self):
        con = duckdb.connect()
        py_obj = "this is a string"
        with pytest.raises(
            duckdb.InvalidInputException, match='Python Object str not suitable to be registered as a view'
        ):
            con.register(py_obj, "v")

    def test_arrow_fetch_table_error(self):
        pytest.importorskip('pyarrow')

        con = duckdb.connect()
        arrow_object = con.execute("select 1").arrow()
        arrow_relation = con.from_arrow(arrow_object)
        res = arrow_relation.execute()
        res.close()
        with pytest.raises(duckdb.InvalidInputException, match='There is no query result'):
            res.fetch_arrow_table()

    def test_arrow_record_batch_reader_error(self):
        pytest.importorskip('pyarrow')

        con = duckdb.connect()
        arrow_object = con.execute("select 1").arrow()
        arrow_relation = con.from_arrow(arrow_object)
        res = arrow_relation.execute()
        res.close()
        with pytest.raises(duckdb.ProgrammingError, match='There is no query result'):
            res.fetch_arrow_reader(1)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_relation_fetchall_error(self, pandas):
        conn = duckdb.connect()
        df_in = pandas.DataFrame(
            {
                'numbers': [1, 2, 3, 4, 5],
            }
        )
        conn.execute("create view x as select * from df_in")
        rel = conn.query("select * from x")
        del df_in
        with pytest.raises(duckdb.ProgrammingError, match='Table with name df_in does not exist'):
            rel.fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_relation_fetchall_execute(self, pandas):
        conn = duckdb.connect()
        df_in = pandas.DataFrame(
            {
                'numbers': [1, 2, 3, 4, 5],
            }
        )
        conn.execute("create view x as select * from df_in")
        rel = conn.query("select * from x")
        del df_in
        with pytest.raises(duckdb.ProgrammingError, match='Table with name df_in does not exist'):
            rel.execute()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_relation_query_error(self, pandas):
        conn = duckdb.connect()
        df_in = pandas.DataFrame(
            {
                'numbers': [1, 2, 3, 4, 5],
            }
        )
        conn.execute("create view x as select * from df_in")
        rel = conn.query("select * from x")
        del df_in
        with pytest.raises(duckdb.CatalogException, match='Table with name df_in does not exist'):
            rel.query("bla", "select * from bla")

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_conn_broken_statement_error(self, pandas):
        conn = duckdb.connect()
        df_in = pandas.DataFrame(
            {
                'numbers': [1, 2, 3, 4, 5],
            }
        )
        conn.execute("create view x as select * from df_in")
        del df_in
        with pytest.raises(duckdb.InvalidInputException):
            conn.execute("select 1; select * from x; select 3;")

    def test_conn_prepared_statement_error(self):
        conn = duckdb.connect()
        conn.execute("create table integers (a integer, b integer)")
        with pytest.raises(duckdb.InvalidInputException, match='Prepared statement needs 2 parameters, 1 given'):
            conn.execute("select * from integers where a =? and b=?", [1])

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_closed_conn_exceptions(self, pandas):
        conn = duckdb.connect()
        conn.close()
        df_in = pandas.DataFrame(
            {
                'numbers': [1, 2, 3, 4, 5],
            }
        )

        with closed():
            conn.register("bla", df_in)

        with closed():
            conn.from_query("select 1")

        with closed():
            conn.table("bla")

        with closed():
            conn.table("bla")

        with closed():
            conn.view("bla")

        with closed():
            conn.values("bla")

        with closed():
            conn.table_function("bla")

        with closed():
            conn.from_df(df_in)

        with closed():
            conn.from_csv_auto("bla")

        with closed():
            conn.from_parquet("bla")

        with closed():
            conn.from_arrow("bla")

    def test_missing_result_from_conn_exceptions(self):
        conn = duckdb.connect()

        with no_result_set():
            conn.fetchone()

        with no_result_set():
            conn.fetchall()

        with no_result_set():
            conn.fetchnumpy()

        with no_result_set():
            conn.fetchdf()

        with no_result_set():
            conn.fetch_df_chunk()

        with no_result_set():
            conn.fetch_arrow_table()

        with no_result_set():
            conn.fetch_record_batch()



from fsspec import filesystem, AbstractFileSystem
from fsspec.implementations.memory import MemoryFileSystem
from .bytes_io_wrapper import BytesIOWrapper
from io import TextIOBase


def is_file_like(obj):
    # We only care that we can read from the file
    return hasattr(obj, "read") and hasattr(obj, "seek")


class ModifiedMemoryFileSystem(MemoryFileSystem):
    protocol = ('DUCKDB_INTERNAL_OBJECTSTORE',)
    # defer to the original implementation that doesn't hardcode the protocol
    _strip_protocol = classmethod(AbstractFileSystem._strip_protocol.__func__)

    # Add this manually because it's apparently missing on windows???
    def unstrip_protocol(self, name):
        """Format FS-specific path to generic, including protocol"""
        protos = (self.protocol,) if isinstance(self.protocol, str) else self.protocol
        for protocol in protos:
            if name.startswith(f"{protocol}://"):
                return name
        return f"{protos[0]}://{name}"

    def info(self, path, **kwargs):
        path = self._strip_protocol(path)
        if path in self.store:
            filelike = self.store[path]
            return {
                "name": path,
                "size": getattr(filelike, "size", 0),
                "type": "file",
                "created": getattr(filelike, "created", None),
            }
        else:
            raise FileNotFoundError(path)

    def _open(
        self,
        path,
        mode="rb",
        block_size=None,
        autocommit=True,
        cache_options=None,
        **kwargs,
    ):
        path = self._strip_protocol(path)
        if path in self.store:
            f = self.store[path]
            return f
        else:
            raise FileNotFoundError(path)

    def add_file(self, object, path):
        if not is_file_like(object):
            raise ValueError("Can not read from a non file-like object")
        path = self._strip_protocol(path)
        if isinstance(object, TextIOBase):
            # Wrap this so that we can return a bytes object from 'read'
            self.store[path] = BytesIOWrapper(object)
        else:
            self.store[path] = object



import pytest
import duckdb
import sys

pd = pytest.importorskip("pandas")
import numpy as np


def compare_results(con, query, expected):
    expected = pd.DataFrame.from_dict(expected)

    unsorted_res = con.query(query).df()
    print(unsorted_res, unsorted_res['a'][0].__class__)
    df_duck = con.query("select * from unsorted_res order by all").df()
    print(df_duck, df_duck['a'][0].__class__)
    print(expected, expected['a'][0].__class__)
    pd.testing.assert_frame_equal(df_duck, expected)


def list_test_cases():
    # fmt: off
    test_cases = [
        ("SELECT list_value(3,5,10) as a", {
            'a': [
                [3, 5, 10]
            ]
        }),
        ("SELECT list_value(3,5,NULL) as a", {
            'a': [
                np.ma.array(
                    [3, 5, 0],
                mask=[0, 0, 1],
            )
            ]
        }),
        ("SELECT list_value(NULL,NULL,NULL) as a", {
            'a': [
                np.ma.array(
                    [0, 0, 0],
                mask=[1, 1, 1],
            )
            ]
        }),
        ("SELECT list_value() as a", {
            'a': [
                np.array([])
            ]
        }),
        ("SELECT list_value('test','test_one','test_two') as a", {
            'a': [
                np.array([
                    'test', 'test_one', 'test_two'
                ])
            ]
        }),
        ("SELECT a from (SELECT LIST(i) as a FROM range(10000) tbl(i)) as t", {
            'a': [
                list(range(0, 10000))
            ]
        }),
        ("SELECT LIST(i) as a FROM range(5) tbl(i) group by i%2 order by all", {
            'a': [
                [0, 2, 4],
                [1, 3]
            ]
        }),
        ("SELECT list_value(1) as a FROM range(5) tbl(i)", {
            'a': [
                [1],
                [1],
                [1],
                [1],
                [1]
            ]
        }),
        ("""
            SELECT
                list(st) AS a
            FROM
            (
                SELECT
                    i,
                    CASE WHEN i%5
                        THEN NULL
                        ELSE i::VARCHAR
                    END AS st
                FROM range(10) tbl(i)
            ) AS t
            GROUP BY i%2
            ORDER BY all
        """, {
            'a': [
                ['0', None, None, None, None],
                [None, None, '5', None, None]
            ]
        }),
    ]

    return test_cases
    # These tests are problematic
    # They cause errors on highest NumPy version supported by 3.7
    # And a crash on manylinux 2014, x86_64 AMD, python 3.12.1
    test_cases.extend([
        ("""
            SELECT * from values
                ([[1, 3], [0,2,4]])
            t(a)
        """, {
            'a': [
                [
                    [1, 3],
                    [0, 2, 4]
                ]
            ]
        }),
        ("""
            SELECT * from values
                ([[[[[0, 2, 4], [1, 3]]]]])
            t(a)
        """, {
            'a': [
                [[[[[0, 2, 4], [1, 3]]]]]
            ]
        }),
        ("""
            select * from values
                ([
                    [NULL],
                    [NULL],
                    [2, 6],
                    [3]
                ])
            t(a)
        """, {
            'a': [
                np.array([
                    np.ma.array([0], mask=[1], dtype=np.int32()),
                    np.ma.array([0], mask=[1], dtype=np.int32()),
                    np.array([2, 6]),
                    np.array([3]),
                ], dtype='object')
            ]
        }),
    ])
    # fmt: on
    return test_cases


class TestFetchNested(object):
    @pytest.mark.parametrize('query, expected', list_test_cases())
    def test_fetch_df_list(self, duckdb_cursor, query, expected):
        compare_results(duckdb_cursor, query, expected)

    # fmt: off
    @pytest.mark.parametrize('query, expected', [
        ("SELECT a from (SELECT STRUCT_PACK(a := 42, b := 43) as a) as t", {
            'a': [
                {'a': 42, 'b': 43}
            ]
        }),
        ("SELECT a from (SELECT STRUCT_PACK(a := NULL) as a) as t", {
            'a': [
                {'a': None}
            ]
        }),
        ("SELECT a from (SELECT STRUCT_PACK(a := i, b := i) as a FROM range(10) tbl(i)) as t", {
            'a': [
                {'a': 0, 'b': 0},
                {'a': 1, 'b': 1},
                {'a': 2, 'b': 2},
                {'a': 3, 'b': 3},
                {'a': 4, 'b': 4},
                {'a': 5, 'b': 5},
                {'a': 6, 'b': 6},
                {'a': 7, 'b': 7},
                {'a': 8, 'b': 8},
                {'a': 9, 'b': 9},
            ]
        }),
        ("SELECT a from (SELECT STRUCT_PACK(a := LIST_VALUE(1,2,3), b := i) as a FROM range(10) tbl(i)) as t", {
            'a': [
                {'a': [1, 2, 3], 'b': 0},
                {'a': [1, 2, 3], 'b': 1},
                {'a': [1, 2, 3], 'b': 2},
                {'a': [1, 2, 3], 'b': 3},
                {'a': [1, 2, 3], 'b': 4},
                {'a': [1, 2, 3], 'b': 5},
                {'a': [1, 2, 3], 'b': 6},
                {'a': [1, 2, 3], 'b': 7},
                {'a': [1, 2, 3], 'b': 8},
                {'a': [1, 2, 3], 'b': 9},
            ]
        }),
    ])
    # fmt: on
    def test_struct_df(self, duckdb_cursor, query, expected):
        compare_results(duckdb_cursor, query, expected)

    # fmt: off
    @pytest.mark.parametrize('query, expected, expected_error', [
        ("SELECT a from (select MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as a) as t", {
            'a': [
                {
                    'key': [1, 2, 3, 4],
                    'value': [10, 9, 8, 7]
                }
            ]
        }, ""),
        ("SELECT a from (select MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as a) as t", {
            'a': [
                {
                    'key': [1, 2, 3, 4],
                    'value': [10, 9, 8, 7]
                }
            ]
        }, ""),
        ("SELECT a from (select MAP(LIST_VALUE(),LIST_VALUE()) as a) as t", {
            'a': [
                {
                    'key': [],
                    'value': []
                }
            ]
        }, ""),
        ("SELECT m as a from (select MAP(list_value(1), list_value(2)) from range(5) tbl(i)) tbl(m)", {
            'a': [
                {'key': [1], 'value': [2]},
                {'key': [1], 'value': [2]},
                {'key': [1], 'value': [2]},
                {'key': [1], 'value': [2]},
                {'key': [1], 'value': [2]},
            ]
        }, ""),
        ("SELECT m as a from (select MAP(lsta,lstb) as m from (SELECT list(i) as lsta, list(i) as lstb from range(10) tbl(i) group by i%5 order by all) as lst_tbl) as T", {
            'a': [
                {'key': [0, 5], 'value': [0, 5]},
                {'key': [1, 6], 'value': [1, 6]},
                {'key': [2, 7], 'value': [2, 7]},
                {'key': [3, 8], 'value': [3, 8]},
                {'key': [4, 9], 'value': [4, 9]},
            ]
        }, ""),
        ("SELECT a from (select MAP(LIST_VALUE(1, 2, 3, 4,2, NULL),LIST_VALUE(10, 9, 8, 7,11,42)) as a) as t", {
            'a': [
                {
                    'key': [1, 2, 3, 4, 2, None],
                    'value': [10, 9, 8, 7, 11, 42]
                }
            ]
        }, "Map keys must be unique"),
        ("SELECT a from (select MAP(LIST_VALUE('Jon Lajoie', 'Backstreet Boys', 'Tenacious D','Jon Lajoie' ),LIST_VALUE(10,9,10,11)) as a) as t", {
            'a': [
                {
                    'key': ['Jon Lajoie', 'Backstreet Boys', 'Tenacious D', 'Jon Lajoie'],
                    'value': [10, 9, 10, 11]
                }
            ]
        }, "Map keys must be unique"),
        ("SELECT a from (select MAP(LIST_VALUE('Jon Lajoie', NULL, 'Tenacious D',NULL,NULL ),LIST_VALUE(10,9,10,11,13)) as a) as t", {
            'a': [
                {
                    'key': ['Jon Lajoie', None, 'Tenacious D', None, None],
                    'value': [10, 9, 10, 11, 13]
                }
            ]
        }, "Map keys can not be NULL"),
        ("SELECT a from (select MAP(LIST_VALUE(NULL, NULL, NULL,NULL,NULL ),LIST_VALUE(10,9,10,11,13)) as a) as t", {
            'a': [
                {
                    'key': [None, None, None, None, None],
                    'value': [10, 9, 10, 11, 13]
                }
            ]
        }, "Map keys can not be NULL"),
        ("SELECT a from (select MAP(LIST_VALUE(NULL, NULL, NULL,NULL,NULL ),LIST_VALUE(NULL, NULL, NULL,NULL,NULL )) as a) as t", {
            'a': [
                {
                    'key': [None, None, None, None, None],
                    'value': [None, None, None, None, None]
                }
            ]
        }, "Map keys can not be NULL"),
    ])
    # fmt: on
    def test_map_df(self, duckdb_cursor, query, expected, expected_error):
        if not expected_error:
            compare_results(duckdb_cursor, query, expected)
        else:
            with pytest.raises(duckdb.InvalidInputException, match=expected_error):
                compare_results(duckdb_cursor, query, expected)

    # fmt: off
    @pytest.mark.parametrize('query, expected', [
        ("""
            SELECT [
                {'i':1,'j':2},
                NULL,
                {'i':2,'j':NULL}
            ] as a
        """, {
            'a': [
                np.ma.array([
                    {'i': 1, 'j': 2},
                    None,
                    {'i': 2, 'j': None}
                ], mask=[0, 1, 0])
            ]
        }),
        ("""
            SELECT [{'i':1,'j':2},NULL,{'i':2,'j':NULL}] as a
        """, {
            'a': [
                np.ma.array([
                    {'i': 1, 'j': 2},
                    None,
                    {'i': 2, 'j': None}
                ], mask=[0, 1, 0])
            ]
        }),
        ("""
            SELECT [{'i':1,'j':[2,3]},NULL] as a
        """, {
            'a': [
                np.ma.array([
                    {'i': 1, 'j': [2, 3]},
                    None,
                ], mask=[0, 1])
            ]
        }),
        ("""
            SELECT {'i':mp,'j':mp2} as a FROM (SELECT MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as mp, MAP(LIST_VALUE(1, 2, 3, 5),LIST_VALUE(10, 9, 8, 7)) as mp2) as t
        """, {
            'a': [
                {
                    'i': {'key': [1, 2, 3, 4], 'value': [10, 9, 8, 7]},
                    'j': {'key': [1, 2, 3, 5], 'value': [10, 9, 8, 7]}
                }
            ]
        }),
        ("""
            SELECT [mp,mp2] as a FROM (SELECT MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as mp, MAP(LIST_VALUE(1, 2, 3, 5),LIST_VALUE(10, 9, 8, 7)) as mp2) as t
        """, {
            'a': [
                [{'key': [1, 2, 3, 4], 'value': [10, 9, 8, 7]}, {'key': [1, 2, 3, 5], 'value': [10, 9, 8, 7]}]
            ]
        }),
        ("""
            SELECT MAP(LIST_VALUE([1,2],[3,4],[5,4]),LIST_VALUE([1,2],[3,4],[5,4])) as a
        """, {
            'a': [
                {'key': [[1, 2], [3, 4], [5, 4]], 'value': [[1, 2], [3, 4], [5, 4]]}
            ]
        }),
        ("""
            SELECT MAP(LIST_VALUE({'i':1,'j':2},{'i':3,'j':4}),LIST_VALUE({'i':1,'j':2},{'i':3,'j':4})) as a
        """, {
            'a': [
                {'key': [{'i': 1, 'j': 2}, {'i': 3, 'j': 4}], 'value': [{'i': 1, 'j': 2}, {'i': 3, 'j': 4}]}
            ]
        }),
        ("""
            SELECT [{'i':1,'j':[2,3]},NULL,{'i':1,'j':[2,3]}] as a
        """, {
            'a': [
                np.ma.array([
                    {'i': 1, 'j': [2, 3]},
                    None,
                    {'i': 1, 'j': [2, 3]}
                ], mask=[0, 1, 0])
            ]
        }),
        ("""
            SELECT * FROM (VALUES
            ({'i':1,'j':2}),
            ({'i':1,'j':2}),
            (NULL),
            (NULL)
        ) t(a)
        """, {
            'a': [
                {'i': 1, 'j': 2},
                {'i': 1, 'j': 2},
                np.nan,
                np.nan
            ]
        }),
        ("""
            SELECT a FROM (VALUES
                (MAP(LIST_VALUE(1,2),LIST_VALUE(3,4))),
                (MAP(LIST_VALUE(1,2),LIST_VALUE(3,4))),
                (NULL),
                (NULL)
            ) t(a)
        """, {
            'a': [
                {'key': [1, 2], 'value': [3, 4]},
                {'key': [1, 2], 'value': [3, 4]},
                np.nan,
                np.nan
            ]
        }),
    ])
    # fmt: on
    def test_nested_mix(self, duckdb_cursor, query, expected):
        compare_results(duckdb_cursor, query, expected)



import numpy as np
import datetime
import duckdb
import pytest


def assert_nullness(items, null_indices):
    for i in range(len(items)):
        if i in null_indices:
            assert items[i] == None
        else:
            assert items[i] != None


class TestPandasNA(object):
    def test_pandas_na(self, duckdb_cursor):
        pd = pytest.importorskip('pandas', minversion='1.0.0', reason='Support for pandas.NA has not been added yet')
        # DataFrame containing a single pd.NA
        df = pd.DataFrame(pd.Series([pd.NA]))

        res = duckdb_cursor.execute("select * from df").fetchall()
        assert res[0][0] == None

        # DataFrame containing multiple values, with a pd.NA mixed in
        null_index = 3
        df = pd.DataFrame(pd.Series([3, 1, 2, pd.NA, 8, 6]))
        res = duckdb_cursor.execute("select * from df").fetchall()
        items = [x[0] for x in [y for y in res]]
        assert_nullness(items, [null_index])

        # Test if pd.NA behaves the same as np.NaN once converted
        nan_df = pd.DataFrame(
            {
                'a': [
                    1.123,
                    5.23234,
                    np.NaN,
                    7234.0000124,
                    0.000000124,
                    0000000000000.0000001,
                    np.NaN,
                    -2342349234.00934580345,
                ]
            }
        )
        na_df = pd.DataFrame(
            {
                'a': [
                    1.123,
                    5.23234,
                    pd.NA,
                    7234.0000124,
                    0.000000124,
                    0000000000000.0000001,
                    pd.NA,
                    -2342349234.00934580345,
                ]
            }
        )
        assert str(nan_df['a'].dtype) == 'float64'
        assert str(na_df['a'].dtype) == 'object'  # pd.NA values turn the column into 'object'

        nan_result = duckdb_cursor.execute("select * from nan_df").df()
        na_result = duckdb_cursor.execute("select * from na_df").df()
        pd.testing.assert_frame_equal(nan_result, na_result)

        # Mixed with stringified pd.NA values
        na_string_df = pd.DataFrame({'a': [str(pd.NA), str(pd.NA), pd.NA, str(pd.NA), pd.NA, pd.NA, pd.NA, str(pd.NA)]})
        null_indices = [2, 4, 5, 6]
        res = duckdb_cursor.execute("select * from na_string_df").fetchall()
        items = [x[0] for x in [y for y in res]]
        assert_nullness(items, null_indices)



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")
from duckdb.experimental.spark.sql.types import Row
from duckdb.experimental.spark.sql.functions import concat_ws, col


class TestReplaceEmpty(object):
    def test_replace_empty(self, spark):
        data = [
            ("firstRowFirstColumn", "firstRowSecondColumn"),
            ("2ndRowFirstColumn", "2ndRowSecondColumn"),
        ]
        df = spark.createDataFrame(data, ["firstColumn", "secondColumn"])
        df = df.withColumn("concatted", concat_ws(" ", col("firstColumn"), col("secondColumn")))
        res = df.select("concatted").collect()
        assert res == [
            Row(concatted="firstRowFirstColumn firstRowSecondColumn"),
            Row(concatted="2ndRowFirstColumn 2ndRowSecondColumn"),
        ]



import duckdb

import pytest

pa = pytest.importorskip("pyarrow")
pq = pytest.importorskip("pyarrow.parquet")
ds = pytest.importorskip("pyarrow.dataset")
np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")
import datetime

Timestamp = pd.Timestamp


class TestArrowDictionary(object):
    def test_dictionary(self, duckdb_cursor):
        indices = pa.array([0, 1, 0, 1, 2, 1, 0, 2])
        dictionary = pa.array([10, 100, None])
        dict_array = pa.DictionaryArray.from_arrays(indices, dictionary)
        arrow_table = pa.Table.from_arrays([dict_array], ['a'])
        rel = duckdb_cursor.from_arrow(arrow_table)

        assert rel.execute().fetchall() == [(10,), (100,), (10,), (100,), (None,), (100,), (10,), (None,)]

        # Bigger than Vector Size
        indices_list = [0, 1, 0, 1, 2, 1, 0, 2, 3] * 10000
        indices = pa.array(indices_list)
        dictionary = pa.array([10, 100, None, 999999])
        dict_array = pa.DictionaryArray.from_arrays(indices, dictionary)
        arrow_table = pa.Table.from_arrays([dict_array], ['a'])
        rel = duckdb_cursor.from_arrow(arrow_table)
        result = [(10,), (100,), (10,), (100,), (None,), (100,), (10,), (None,), (999999,)] * 10000
        assert rel.execute().fetchall() == result

        # Table with dictionary and normal array

        arrow_table = pa.Table.from_arrays([dict_array, pa.array(indices_list)], ['a', 'b'])
        rel = duckdb_cursor.from_arrow(arrow_table)
        result = [(10, 0), (100, 1), (10, 0), (100, 1), (None, 2), (100, 1), (10, 0), (None, 2), (999999, 3)] * 10000
        assert rel.execute().fetchall() == result

    def test_dictionary_null_index(self, duckdb_cursor):
        indices = pa.array([None, 1, 0, 1, 2, 1, 0, 2])
        dictionary = pa.array([10, 100, None])
        dict_array = pa.DictionaryArray.from_arrays(indices, dictionary)
        arrow_table = pa.Table.from_arrays([dict_array], ['a'])
        rel = duckdb_cursor.from_arrow(arrow_table)

        assert rel.execute().fetchall() == [(None,), (100,), (10,), (100,), (None,), (100,), (10,), (None,)]

        indices = pa.array([None, 1, None, 1, 2, 1, 0])
        dictionary = pa.array([10, 100, 100])
        dict_array = pa.DictionaryArray.from_arrays(indices, dictionary)
        arrow_table = pa.Table.from_arrays([dict_array], ['a'])
        rel = duckdb_cursor.from_arrow(arrow_table)
        print(rel.execute().fetchall())
        assert rel.execute().fetchall() == [(None,), (100,), (None,), (100,), (100,), (100,), (10,)]

        # Test Big Vector
        indices_list = [None, 1, None, 1, 2, 1, 0]
        indices = pa.array(indices_list * 1000)
        dictionary = pa.array([10, 100, 100])
        dict_array = pa.DictionaryArray.from_arrays(indices, dictionary)
        arrow_table = pa.Table.from_arrays([dict_array], ['a'])
        rel = duckdb_cursor.from_arrow(arrow_table)
        result = [(None,), (100,), (None,), (100,), (100,), (100,), (10,)] * 1000
        assert rel.execute().fetchall() == result

        # Table with dictionary and normal array
        arrow_table = pa.Table.from_arrays([dict_array, indices], ['a', 'b'])
        rel = duckdb_cursor.from_arrow(arrow_table)
        result = [(None, None), (100, 1), (None, None), (100, 1), (100, 2), (100, 1), (10, 0)] * 1000
        assert rel.execute().fetchall() == result

    @pytest.mark.parametrize(
        'element',
        [
            # list
            """
            ['hello'::ENUM('hello', 'bye')]
        """,
            # struct
            """
            {'a': 'hello'::ENUM('hello', 'bye')}
        """,
            # union
            """
            {'a': 'hello'::ENUM('hello', 'bye')}::UNION(a integer, b bool, c struct(a enum('hello', 'bye')))
        """,
            # map (key)
            """
            map {'hello'::ENUM('hello', 'bye') : 'test'}
        """,
            # map (val)
            """
            map {'test': 'hello'::ENUM('hello', 'bye')}
        """,
            # list of struct(enum)
            """
            [{'a': 'hello'::ENUM('hello', 'bye')}]
        """,
            # list of union(enum)
            """
            [{'a': 'hello'::ENUM('hello', 'bye')}::UNION(a integer, b bool, c struct(a enum('hello', 'bye')))]
        """,
            # list of list
            """
            [['hello'::ENUM('hello', 'bye')], [], NULL, ['hello'::ENUM('hello', 'bye'), 'bye'::ENUM('hello', 'bye')]]
        """,
        ],
    )
    @pytest.mark.parametrize(
        'count',
        [
            1,
            10,
            1024,
            # 2048,
            # 2047,
            # 2049,
            # 4000,
            # 4096,
            5000,
        ],
    )
    @pytest.mark.parametrize('query', ["select {} as a from range({})", "select [{} for x in range({})] as a"])
    def test_dictionary_roundtrip(self, query, element, duckdb_cursor, count):
        query = query.format(element, count)
        original_rel = duckdb_cursor.sql(query)
        expected = original_rel.fetchall()
        arrow_res = original_rel.arrow()

        roundtrip_rel = duckdb_cursor.sql('select * from arrow_res')
        actual = roundtrip_rel.fetchall()
        assert expected == actual
        assert original_rel.columns == roundtrip_rel.columns
        # Note: we can't check the types, because originally these are ENUM
        # but because the dictionary of the ENUM can not be known before execution we output VARCHAR instead.

    def test_dictionary_batches(self, duckdb_cursor):
        indices_list = [None, 1, None, 1, 2, 1, 0]
        indices = pa.array(indices_list * 10000)
        dictionary = pa.array([10, 100, 100])
        dict_array = pa.DictionaryArray.from_arrays(indices, dictionary)
        arrow_table = pa.Table.from_arrays([dict_array], ['a'])
        batch_arrow_table = pa.Table.from_batches(arrow_table.to_batches(10))
        rel = duckdb_cursor.from_arrow(batch_arrow_table)
        result = [(None,), (100,), (None,), (100,), (100,), (100,), (10,)] * 10000
        assert rel.execute().fetchall() == result

        # Table with dictionary and normal array
        arrow_table = pa.Table.from_arrays([dict_array, indices], ['a', 'b'])
        batch_arrow_table = pa.Table.from_batches(arrow_table.to_batches(10))
        rel = duckdb_cursor.from_arrow(batch_arrow_table)
        result = [(None, None), (100, 1), (None, None), (100, 1), (100, 2), (100, 1), (10, 0)] * 10000
        assert rel.execute().fetchall() == result

    def test_dictionary_lifetime(self, duckdb_cursor):
        tables = []
        expected = ''
        for i in range(100):
            if i % 3 == 0:
                input = 'ABCD' * 17000
            elif i % 3 == 1:
                input = 'FOOO' * 17000
            else:
                input = 'BARR' * 17000
            expected += input
            array = pa.array(
                input,
                type=pa.dictionary(pa.int16(), pa.string()),
            )
            tables.append(pa.table([array], names=["x"]))
        # All of the tables with different dictionaries are getting merged into one dataset
        # This is testing that our cache is being evicted correctly
        x = ds.dataset(tables)
        res = duckdb_cursor.sql("select * from x").fetchall()
        expected = [(x,) for x in expected]
        assert res == expected

    def test_dictionary_batches_parallel(self, duckdb_cursor):
        duckdb_cursor.execute("PRAGMA threads=4")
        duckdb_cursor.execute("PRAGMA verify_parallelism")

        indices_list = [None, 1, None, 1, 2, 1, 0]
        indices = pa.array(indices_list * 10000)
        dictionary = pa.array([10, 100, 100])
        dict_array = pa.DictionaryArray.from_arrays(indices, dictionary)
        arrow_table = pa.Table.from_arrays([dict_array], ['a'])
        batch_arrow_table = pa.Table.from_batches(arrow_table.to_batches(10))
        rel = duckdb_cursor.from_arrow(batch_arrow_table)
        result = [(None,), (100,), (None,), (100,), (100,), (100,), (10,)] * 10000
        assert rel.execute().fetchall() == result

        # Table with dictionary and normal array
        arrow_table = pa.Table.from_arrays([dict_array, indices], ['a', 'b'])
        batch_arrow_table = pa.Table.from_batches(arrow_table.to_batches(10))
        rel = duckdb_cursor.from_arrow(batch_arrow_table)
        result = [(None, None), (100, 1), (None, None), (100, 1), (100, 2), (100, 1), (10, 0)] * 10000
        assert rel.execute().fetchall() == result

    def test_dictionary_index_types(self, duckdb_cursor):
        indices_list = [None, 1, None, 1, 2, 1, 0]
        dictionary = pa.array([10, 100, 100], type=pa.uint8())
        index_types = []
        index_types.append(pa.array(indices_list * 10000, type=pa.uint8()))
        index_types.append(pa.array(indices_list * 10000, type=pa.uint16()))
        index_types.append(pa.array(indices_list * 10000, type=pa.uint32()))
        index_types.append(pa.array(indices_list * 10000, type=pa.uint64()))
        index_types.append(pa.array(indices_list * 10000, type=pa.int8()))
        index_types.append(pa.array(indices_list * 10000, type=pa.int16()))
        index_types.append(pa.array(indices_list * 10000, type=pa.int32()))
        index_types.append(pa.array(indices_list * 10000, type=pa.int64()))

        for index_type in index_types:
            dict_array = pa.DictionaryArray.from_arrays(index_type, dictionary)
            arrow_table = pa.Table.from_arrays([dict_array], ['a'])
            rel = duckdb_cursor.from_arrow(arrow_table)
            result = [(None,), (100,), (None,), (100,), (100,), (100,), (10,)] * 10000
            assert rel.execute().fetchall() == result

    def test_dictionary_strings(self, duckdb_cursor):
        indices_list = [None, 0, 1, 2, 3, 4, None]
        indices = pa.array(indices_list * 1000)
        dictionary = pa.array(['Matt Daaaaaaaaamon', 'Alec Baldwin', 'Sean Penn', 'Tim Robbins', 'Samuel L. Jackson'])
        dict_array = pa.DictionaryArray.from_arrays(indices, dictionary)
        arrow_table = pa.Table.from_arrays([dict_array], ['a'])
        rel = duckdb_cursor.from_arrow(arrow_table)
        result = [
            (None,),
            ('Matt Daaaaaaaaamon',),
            ('Alec Baldwin',),
            ('Sean Penn',),
            ('Tim Robbins',),
            ('Samuel L. Jackson',),
            (None,),
        ] * 1000
        assert rel.execute().fetchall() == result

    def test_dictionary_timestamps(self, duckdb_cursor):
        indices_list = [None, 0, 1, 2, None]
        indices = pa.array(indices_list * 1000)
        dictionary = pa.array(
            [
                Timestamp(year=2001, month=9, day=25),
                Timestamp(year=2006, month=11, day=14),
                Timestamp(year=2012, month=5, day=15),
                Timestamp(year=2018, month=11, day=2),
            ]
        )
        dict_array = pa.DictionaryArray.from_arrays(indices, dictionary)
        arrow_table = pa.Table.from_arrays([dict_array], ['a'])
        rel = duckdb_cursor.from_arrow(arrow_table)
        print(rel.execute().fetchall())
        expected = [
            (None,),
            (datetime.datetime(2001, 9, 25, 0, 0),),
            (datetime.datetime(2006, 11, 14, 0, 0),),
            (datetime.datetime(2012, 5, 15, 0, 0),),
            (None,),
        ] * 1000
        result = rel.execute().fetchall()
        assert result == expected



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")
from duckdb.experimental.spark.sql.catalog import Table, Database, Column


class TestSparkCatalog(object):
    def test_list_databases(self, spark):
        dbs = spark.catalog.listDatabases()
        assert dbs == [
            Database(name='memory', description=None, locationUri=''),
            Database(name='system', description=None, locationUri=''),
            Database(name='temp', description=None, locationUri=''),
        ]

    def test_list_tables(self, spark):
        # empty
        tbls = spark.catalog.listTables()
        assert tbls == []

        spark.sql('create table tbl(a varchar)')
        tbls = spark.catalog.listTables()
        assert tbls == [
            Table(
                name='tbl',
                database='memory',
                description='CREATE TABLE tbl(a VARCHAR);',
                tableType='',
                isTemporary=False,
            )
        ]

    def test_list_columns(self, spark):
        spark.sql('create table tbl(a varchar, b bool)')
        columns = spark.catalog.listColumns('tbl')
        assert columns == [
            Column(name='a', description=None, dataType='VARCHAR', nullable=True, isPartition=False, isBucket=False),
            Column(name='b', description=None, dataType='BOOLEAN', nullable=True, isPartition=False, isBucket=False),
        ]

        # FIXME: should this error instead?
        non_existant_columns = spark.catalog.listColumns('none_existant')
        assert non_existant_columns == []

        spark.sql('create view vw as select * from tbl')
        view_columns = spark.catalog.listColumns('vw')
        assert view_columns == columns



import duckdb
import pandas as pd
import pytest


class TestAmbiguousPrepare(object):
    def test_bool(self, duckdb_cursor):
        conn = duckdb.connect()
        res = conn.execute("select ?, ?, ?", (True, 42, [1, 2, 3])).fetchall()
        assert res[0][0] == True
        assert res[0][1] == 42
        assert res[0][2] == [1, 2, 3]



import duckdb
import pytest
import string
import datetime as dt

pa = pytest.importorskip("pyarrow")


# Reconstruct filters when pushing down into arrow scan
# arrow supports timestamp_tz with different units than US, we only support US
# so we have to convert ConstantValues back to their native unit when pushing the filter expression containing them down to pyarrow
class Test8522(object):
    def test_8522(self, duckdb_cursor):
        t_us = pa.Table.from_arrays(
            arrays=[pa.array([dt.datetime(2022, 1, 1)])],
            schema=pa.schema([pa.field("time", pa.timestamp("us", tz="UTC"))]),
        )

        t_ms = pa.Table.from_arrays(
            arrays=[pa.array([dt.datetime(2022, 1, 1)])],
            schema=pa.schema([pa.field("time", pa.timestamp("ms", tz="UTC"))]),
        )

        expected = duckdb_cursor.sql("FROM t_us").filter("time>='2022-01-01'").fetchall()
        assert len(expected) == 1

        actual = duckdb_cursor.sql("FROM t_ms").filter("time>='2022-01-01'").fetchall()
        assert actual == expected



import duckdb
import datetime
import pytz
import os
import pytest

pd = pytest.importorskip("pandas")
pa = pytest.importorskip("pyarrow")
from packaging.version import Version

filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'data', 'tz.parquet')


class TestNativeTimeZone(object):
    def test_native_python_timestamp_timezone(self, duckdb_cursor):
        duckdb_cursor.execute("SET timezone='America/Los_Angeles';")
        res = duckdb_cursor.execute(f"select TimeRecStart as tz  from '{filename}'").fetchone()
        assert res[0].hour == 14 and res[0].minute == 52
        assert res[0].tzinfo.zone == 'America/Los_Angeles'

        res = duckdb_cursor.execute(f"select TimeRecStart as tz  from '{filename}'").fetchall()[0]
        assert res[0].hour == 14 and res[0].minute == 52
        assert res[0].tzinfo.zone == 'America/Los_Angeles'

        res = duckdb_cursor.execute(f"select TimeRecStart as tz  from '{filename}'").fetchmany(1)[0]
        assert res[0].hour == 14 and res[0].minute == 52
        assert res[0].tzinfo.zone == 'America/Los_Angeles'

        duckdb_cursor.execute("SET timezone='UTC';")
        res = duckdb_cursor.execute(f"select TimeRecStart as tz  from '{filename}'").fetchone()
        assert res[0].hour == 21 and res[0].minute == 52
        assert res[0].tzinfo.zone == 'UTC'

    def test_native_python_time_timezone(self, duckdb_cursor):
        res = duckdb_cursor.execute(f"select TimeRecStart::TIMETZ as tz from '{filename}'").fetchone()
        assert res == (datetime.time(21, 52, 27, tzinfo=datetime.timezone.utc),)

        roundtrip = duckdb_cursor.execute("select $1", [res[0]]).fetchone()
        assert roundtrip == res

    def test_pandas_timestamp_timezone(self, duckdb_cursor):
        res = duckdb_cursor.execute("SET timezone='America/Los_Angeles';")
        res = duckdb_cursor.execute(f"select TimeRecStart as tz  from '{filename}'").df()
        assert res.dtypes["tz"].tz.zone == 'America/Los_Angeles'
        assert res['tz'][0].hour == 14 and res['tz'][0].minute == 52

        duckdb_cursor.execute("SET timezone='UTC';")
        res = duckdb_cursor.execute(f"select TimeRecStart as tz  from '{filename}'").df()
        assert res['tz'][0].hour == 21 and res['tz'][0].minute == 52

    def test_pandas_timestamp_time(self, duckdb_cursor):
        with pytest.raises(
            duckdb.NotImplementedException, match="Not implemented Error: Unsupported type \"TIME WITH TIME ZONE\""
        ):
            duckdb_cursor.execute(f"select TimeRecStart::TIMETZ  as tz  from '{filename}'").df()

    @pytest.mark.skipif(
        Version(pa.__version__) < Version('15.0.0'), reason="pyarrow 14.0.2 'to_pandas' causes a DeprecationWarning"
    )
    def test_arrow_timestamp_timezone(self, duckdb_cursor):
        res = duckdb_cursor.execute("SET timezone='America/Los_Angeles';")
        table = duckdb_cursor.execute(f"select TimeRecStart as tz  from '{filename}'").arrow()
        res = table.to_pandas()
        assert res.dtypes["tz"].tz.zone == 'America/Los_Angeles'
        assert res['tz'][0].hour == 14 and res['tz'][0].minute == 52

        duckdb_cursor.execute("SET timezone='UTC';")
        res = duckdb_cursor.execute(f"select TimeRecStart as tz  from '{filename}'").arrow().to_pandas()
        assert res.dtypes["tz"].tz.zone == 'UTC'
        assert res['tz'][0].hour == 21 and res['tz'][0].minute == 52

    def test_arrow_timestamp_time(self, duckdb_cursor):
        with pytest.raises(duckdb.NotImplementedException, match="Unsupported Arrow type"):
            duckdb_cursor.execute(f"select TimeRecStart::TIMETZ  as tz  from '{filename}'").arrow()



# simple DB API testcase

import duckdb
import numpy
import pytest
import re
from conftest import NumpyPandas, ArrowPandas


class TestDBConfig(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_default_order(self, duckdb_cursor, pandas):
        df = pandas.DataFrame({'a': [1, 2, 3]})
        con = duckdb.connect(':memory:', config={'default_order': 'desc'})
        result = con.execute('select * from df order by a').fetchall()
        assert result == [(3,), (2,), (1,)]

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_null_order(self, duckdb_cursor, pandas):
        df = pandas.DataFrame({'a': [1, 2, 3, None]})
        con = duckdb.connect(':memory:', config={'default_null_order': 'nulls_last'})
        result = con.execute('select * from df order by a').fetchall()
        assert result == [(1,), (2,), (3,), (None,)]

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_multiple_options(self, duckdb_cursor, pandas):
        df = pandas.DataFrame({'a': [1, 2, 3, None]})
        con = duckdb.connect(':memory:', config={'default_null_order': 'nulls_last', 'default_order': 'desc'})
        result = con.execute('select * from df order by a').fetchall()
        assert result == [(3,), (2,), (1,), (None,)]

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_external_access(self, duckdb_cursor, pandas):
        df = pandas.DataFrame({'a': [1, 2, 3]})
        # this works (replacement scan)
        con_regular = duckdb.connect(':memory:', config={})
        con_regular.execute('select * from df')
        # disable external access: this also disables pandas replacement scans
        con = duckdb.connect(':memory:', config={'enable_external_access': False})
        # this should fail
        query_failed = False
        try:
            con.execute('select * from df').fetchall()
        except:
            query_failed = True
        assert query_failed == True

    def test_unrecognized_option(self, duckdb_cursor):
        success = True
        try:
            con_regular = duckdb.connect(':memory:', config={'thisoptionisprobablynotthere': '42'})
        except:
            success = False
        assert success == False

    def test_incorrect_parameter(self, duckdb_cursor):
        success = True
        try:
            con_regular = duckdb.connect(':memory:', config={'default_null_order': '42'})
        except:
            success = False
        assert success == False

    def test_user_agent_default(self, duckdb_cursor):
        con_regular = duckdb.connect(':memory:')
        regex = re.compile("duckdb/.* python")
        # Expands to: SELECT * FROM pragma_user_agent()
        assert regex.match(con_regular.sql("PRAGMA user_agent").fetchone()[0]) is not None
        custom_user_agent = con_regular.sql("SELECT current_setting('custom_user_agent')").fetchone()
        assert custom_user_agent[0] == ''

    def test_user_agent_custom(self, duckdb_cursor):
        con_regular = duckdb.connect(':memory:', config={'custom_user_agent': 'CUSTOM_STRING'})
        regex = re.compile("duckdb/.* python CUSTOM_STRING")
        assert regex.match(con_regular.sql("PRAGMA user_agent").fetchone()[0]) is not None
        custom_user_agent = con_regular.sql("SELECT current_setting('custom_user_agent')").fetchone()
        assert custom_user_agent[0] == 'CUSTOM_STRING'

    def test_secret_manager_option(self, duckdb_cursor):
        con = duckdb.connect(':memory:', config={'allow_persistent_secrets': False})
        result = con.execute('select count(*) from duckdb_secrets()').fetchall()
        assert result == [(0,)]



import duckdb
import pytest
import string

pa = pytest.importorskip("pyarrow")
pq = pytest.importorskip("pyarrow.parquet")
pl = pytest.importorskip("polars")


class Test7699(object):
    def test_7699(self, duckdb_cursor):
        pl_tbl = pl.DataFrame(
            {
                "col1": pl.Series([string.ascii_uppercase[ix + 10] for ix in list(range(2)) + list(range(3))]).cast(
                    pl.Categorical
                ),
            }
        )

        nickname = "df1234"
        duckdb_cursor.register(nickname, pl_tbl)

        rel = duckdb_cursor.sql("select * from df1234")
        res = rel.fetchall()
        assert res == [('K',), ('L',), ('K',), ('L',), ('M',)]



import duckdb


class TestContextManager(object):
    def test_context_manager(self, duckdb_cursor):
        with duckdb.connect(database=':memory:', read_only=False) as con:
            assert con.execute("select 1").fetchall() == [(1,)]



import duckdb


class TestMap(object):

    def test_array_list_tuple_ambiguity(self):
        con = duckdb.connect()
        res = con.sql("SELECT $arg", params={'arg': (1, 2)}).fetchall()[0][0]
        assert res == [1, 2]

        # By using an explicit duckdb.Value with an array type, we should convert the input as an array
        # and get an array (tuple) back
        typ = duckdb.array_type(duckdb.typing.BIGINT, 2)
        val = duckdb.Value((1, 2), typ)
        res = con.sql("SELECT $arg", params={'arg': val}).fetchall()[0][0]
        assert res == (1, 2)

        val = duckdb.Value([3, 4], typ)
        res = con.sql("SELECT $arg", params={'arg': val}).fetchall()[0][0]
        assert res == (3, 4)



import duckdb
import pytest


class TestDuckDBExecute(object):
    def test_execute_basic(self, duckdb_cursor):
        duckdb_cursor.execute('create table t as select 5')
        res = duckdb_cursor.table('t').fetchall()
        assert res == [(5,)]

    def test_execute_many_basic(self, duckdb_cursor):
        duckdb_cursor.execute("create table t(x int);")

        # This works because prepared parameter is only present in the last statement
        duckdb_cursor.execute(
            """
			delete from t where x=5;
			insert into t(x) values($1);
		""",
            (99,),
        )
        res = duckdb_cursor.table('t').fetchall()
        assert res == [(99,)]

    def test_execute_many_error(self, duckdb_cursor):
        duckdb_cursor.execute("create table t(x int);")

        # Prepared parameter used in a statement that is not the last
        with pytest.raises(
            duckdb.NotImplementedException, match='Prepared parameters are only supported for the last statement'
        ):
            duckdb_cursor.execute(
                """
				delete from t where x=$1;
				insert into t(x) values($1);
			""",
                (99,),
            )



import sys

import pytest

pa = pytest.importorskip("pyarrow")
adbc_driver_manager = pytest.importorskip("adbc_driver_manager")

adbc_driver_duckdb = pytest.importorskip("adbc_driver_duckdb.dbapi")
con = adbc_driver_duckdb.connect()


def _import(handle):
    """Helper to import a C Data Interface handle."""
    if isinstance(handle, adbc_driver_manager.ArrowArrayStreamHandle):
        return pa.RecordBatchReader._import_from_c(handle.address)
    elif isinstance(handle, adbc_driver_manager.ArrowSchemaHandle):
        return pa.Schema._import_from_c(handle.address)
    raise NotImplementedError(f"Importing {handle!r}")


def _bind(stmt, batch):
    array = adbc_driver_manager.ArrowArrayHandle()
    schema = adbc_driver_manager.ArrowSchemaHandle()
    batch._export_to_c(array.address, schema.address)
    stmt.bind(array, schema)


class TestADBCStatementBind(object):
    def test_bind_multiple_rows(self):
        data = pa.record_batch(
            [
                [1, 2, 3, 4],
            ],
            names=["ints"],
        )

        con = adbc_driver_duckdb.connect()
        with con.cursor() as cursor:
            statement = cursor.adbc_statement
            statement.set_sql_query("select ? * 2 as i")
            statement.prepare()

            _bind(statement, data)
            with pytest.raises(
                adbc_driver_manager.NotSupportedError, match="Binding multiple rows at once is not supported yet"
            ):
                res, number_of_rows = statement.execute_query()

    def test_bind_single_row(self):
        expected_result = pa.array([8], type=pa.int64())

        data = pa.record_batch(
            [
                [4],
            ],
            names=["ints"],
        )

        con = adbc_driver_duckdb.connect()
        with con.cursor() as cursor:
            statement = cursor.adbc_statement
            statement.set_sql_query("select ? * 2 as i")
            statement.prepare()

            raw_schema = statement.get_parameter_schema()
            schema = _import(raw_schema)
            assert schema.names == ['0']

            _bind(statement, data)
            res, _ = statement.execute_query()
            table = _import(res).read_all()

            result = table['i']
            assert result.num_chunks == 1
            result_values = result.chunk(0)
            assert result_values == expected_result

    def test_multiple_parameters(self):
        int_data = pa.array([5])
        varchar_data = pa.array(['not a short string'])
        bool_data = pa.array([True])

        # Create the schema
        schema = pa.schema([('a', pa.int64()), ('b', pa.string()), ('c', pa.bool_())])

        # Create the PyArrow table
        expected_res = pa.Table.from_arrays([int_data, varchar_data, bool_data], schema=schema)

        data = pa.record_batch(
            [[5], ['not a short string'], [True]],
            names=["ints", "strings", "bools"],
        )

        con = adbc_driver_duckdb.connect()
        with con.cursor() as cursor:
            statement = cursor.adbc_statement
            statement.set_sql_query("select ? as a, ? as b, ? as c")
            statement.prepare()

            raw_schema = statement.get_parameter_schema()
            schema = _import(raw_schema)
            assert schema.names == ['0', '1', '2']

            _bind(statement, data)
            res, _ = statement.execute_query()
            table = _import(res).read_all()

            assert table == expected_res

    def test_bind_composite_type(self):
        data_dict = {
            'field1': pa.array([10], type=pa.int64()),
            'field2': pa.array([3.14], type=pa.float64()),
            'field3': pa.array(['example with long string'], type=pa.string()),
        }
        # Create the StructArray
        struct_array = pa.StructArray.from_arrays(arrays=data_dict.values(), names=data_dict.keys())

        schema = pa.schema([(name, array.type) for name, array in zip(['a'], [struct_array])])

        # Create the RecordBatch
        record_batch = pa.RecordBatch.from_arrays([struct_array], schema=schema)

        con = adbc_driver_duckdb.connect()
        with con.cursor() as cursor:
            statement = cursor.adbc_statement
            statement.set_sql_query("select ? as a")
            statement.prepare()

            raw_schema = statement.get_parameter_schema()
            schema = _import(raw_schema)
            assert schema.names == ['0']

            _bind(statement, record_batch)
            res, _ = statement.execute_query()
            table = _import(res).read_all()
            result = table['a']
            result = result.chunk(0)
            assert result == struct_array

    def test_too_many_parameters(self):
        data = pa.record_batch(
            [[12423], ['not a short string']],
            names=["ints", "strings"],
        )

        con = adbc_driver_duckdb.connect()
        with con.cursor() as cursor:
            statement = cursor.adbc_statement
            statement.set_sql_query("select ? as a")
            statement.prepare()

            raw_schema = statement.get_parameter_schema()
            schema = _import(raw_schema)
            assert schema.names == ['0']

            array = adbc_driver_manager.ArrowArrayHandle()
            schema = adbc_driver_manager.ArrowSchemaHandle()

            data._export_to_c(array.address, schema.address)
            statement.bind(array, schema)

            with pytest.raises(
                adbc_driver_manager.ProgrammingError,
                match="Input data has more column than prepared statement has parameters",
            ):
                res, _ = statement.execute_query()

    def test_not_enough_parameters(self):
        data = pa.record_batch(
            [['not a short string']],
            names=["strings"],
        )

        con = adbc_driver_duckdb.connect()
        with con.cursor() as cursor:
            statement = cursor.adbc_statement
            statement.set_sql_query("select ? as a, ? as b")
            statement.prepare()

            raw_schema = statement.get_parameter_schema()
            schema = _import(raw_schema)
            assert schema.names == ['0', '1']

            array = adbc_driver_manager.ArrowArrayHandle()
            schema = adbc_driver_manager.ArrowSchemaHandle()
            data._export_to_c(array.address, schema.address)
            statement.bind(array, schema)
            with pytest.raises(
                adbc_driver_manager.ProgrammingError,
                match="Values were not provided for the following prepared statement parameters: 2",
            ):
                res, _ = statement.execute_query()



import duckdb
import pytest
import pandas as pd
import duckdb

pa = pytest.importorskip("pyarrow", '14.0.0', reason="Needs pyarrow >= 14")
pc = pytest.importorskip("pyarrow.compute")


class TestArrowREE(object):
    @pytest.mark.parametrize(
        'query',
        [
            """
            select
                CASE WHEN (i % 2 == 0) THEN NULL ELSE (i // {})::{} END as ree
            from range({}) t(i);
        """,
            """
            select
                (i // {})::{} as ree
            from range({}) t(i);
        """,
        ],
    )
    @pytest.mark.parametrize('run_length', [4, 1, 10, 1000, 2048, 3000])
    @pytest.mark.parametrize('size', [100, 10000])
    @pytest.mark.parametrize(
        'value_type',
        ['UTINYINT', 'USMALLINT', 'UINTEGER', 'UBIGINT', 'TINYINT', 'SMALLINT', 'INTEGER', 'BIGINT', 'HUGEINT'],
    )
    def test_arrow_run_end_encoding_numerics(self, duckdb_cursor, query, run_length, size, value_type):
        if value_type == 'UTINYINT':
            if size > 255:
                size = 255
        if value_type == 'TINYINT':
            if size > 127:
                size = 127
        query = query.format(run_length, value_type, size)
        rel = duckdb_cursor.sql(query)
        array = rel.arrow()['ree']
        expected = rel.fetchall()

        encoded_array = pc.run_end_encode(array)

        schema = pa.schema([("ree", encoded_array.type)])
        tbl = pa.Table.from_arrays([encoded_array], schema=schema)
        res = duckdb_cursor.sql("select * from tbl").fetchall()
        assert res == expected

    @pytest.mark.parametrize(
        ['dbtype', 'val1', 'val2'],
        [
            ('TINYINT', '(-128)', '127'),
            ('SMALLINT', '(-32768)', '32767'),
            ('INTEGER', '(-2147483648)', '2147483647'),
            ('BIGINT', '(-9223372036854775808)', '9223372036854775807'),
            ('UTINYINT', '0', '255'),
            ('USMALLINT', '0', '65535'),
            ('UINTEGER', '0', '4294967295'),
            ('UBIGINT', '0', '18446744073709551615'),
            ('BOOL', 'true', 'false'),
            ('VARCHAR', "'test'", "'this is a long string'"),
            ('BLOB', "'\\xE0\\x9F\\x98\\x84'", "'\\xF0\\x9F\\xA6\\x86'"),
            ('DATE', "'1992-03-27'", "'2204-11-01'"),
            ('TIME', "'01:02:03'", "'23:41:35'"),
            ('TIMESTAMP_S', "'1992-03-22 01:02:03'", "'2022-11-07 08:43:04.123456'"),
            ('TIMESTAMP', "'1992-03-22 01:02:03'", "'2022-11-07 08:43:04.123456'"),
            ('TIMESTAMP_MS', "'1992-03-22 01:02:03'", "'2022-11-07 08:43:04.123456'"),
            ('TIMESTAMP_NS', "'1992-03-22 01:02:03'", "'2022-11-07 08:43:04.123456'"),
            ('DECIMAL(4,2)', "'12.23'", "'99.99'"),
            ('DECIMAL(7,6)', "'1.234234'", "'0.000001'"),
            ('DECIMAL(14,7)', "'134523.234234'", "'999999.000001'"),
            ('DECIMAL(28,1)', "'12345678910111234123456789.1'", "'999999999999999999999999999.9'"),
            ('UUID', "'10acd298-15d7-417c-8b59-eabb5a2bacab'", "'eeccb8c5-9943-b2bb-bb5e-222f4e14b687'"),
            ('BIT', "'01010101010000'", "'01010100010101010101010101111111111'"),
        ],
    )
    @pytest.mark.parametrize(
        "filter",
        [
            # "ree::VARCHAR == '5'",
            "true"
        ],
    )
    def test_arrow_run_end_encoding(self, duckdb_cursor, dbtype, val1, val2, filter):
        if dbtype in ['BIT', 'UUID']:
            pytest.skip("BIT and UUID are currently broken (FIXME)")
        projection = "a, b, ree"
        query = """
            create table ree_tbl as select
                case when ((i // 8) % 2 == 0)
                    then (
                        case when ((i // 4) % 2 == 0)
                            then {}::{}
                            else {}::{}
                        end
                    ) else
                        NULL
                end as ree,
                i as a,
                i // 5 as b
            from range({}) t(i);
        """
        # Create the table
        query = query.format(val1, dbtype, val2, dbtype, 10000)
        duckdb_cursor.execute(query)

        rel = duckdb_cursor.query("select * from ree_tbl")
        expected = duckdb_cursor.query("select {} from ree_tbl where {}".format(projection, filter)).fetchall()

        # Create an Arrow Table from the table
        arrow_conversion = rel.arrow()
        arrays = {
            'ree': arrow_conversion['ree'],
            'a': arrow_conversion['a'],
            'b': arrow_conversion['b'],
        }

        encoded_arrays = {
            'ree': pc.run_end_encode(arrays['ree']),
            'a': pc.run_end_encode(arrays['a']),
            'b': pc.run_end_encode(arrays['b']),
        }

        schema = pa.schema(
            [
                ("ree", encoded_arrays['ree'].type),
                ("a", encoded_arrays['a'].type),
                ("b", encoded_arrays['b'].type),
            ]
        )
        tbl = pa.Table.from_arrays([encoded_arrays['ree'], encoded_arrays['a'], encoded_arrays['b']], schema=schema)

        # Scan the Arrow Table and verify that the results are the same
        res = duckdb_cursor.sql("select {} from tbl where {}".format(projection, filter)).fetchall()
        assert res == expected

    def test_arrow_ree_empty_table(self, duckdb_cursor):
        duckdb_cursor.query("create table tbl (ree integer)")
        rel = duckdb_cursor.table('tbl')
        array = rel.arrow()['ree']
        expected = rel.fetchall()

        encoded_array = pc.run_end_encode(array)

        schema = pa.schema([("ree", encoded_array.type)])
        pa_res = pa.Table.from_arrays([encoded_array], schema=schema)
        res = duckdb_cursor.sql("select * from pa_res").fetchall()
        assert res == expected

    @pytest.mark.parametrize('projection', ['*', 'a, c, b', 'ree, a, b, c', 'c, b, a, ree', 'c', 'b, ree, c, a'])
    def test_arrow_ree_projections(self, duckdb_cursor, projection):
        # Create the schema
        duckdb_cursor.query(
            """
            create table tbl (
                ree integer,
                a int,
                b bool,
                c varchar
            )
        """
        )

        # Populate the table with data
        duckdb_cursor.query(
            """
            insert into tbl select
                i // 4,
                i,
                i % 2 == 0,
                i::VARCHAR
            from range(100000) t(i)
        """
        )

        # Fetch the result as an Arrow Table
        result = duckdb_cursor.table('tbl').arrow()

        # Turn 'ree' into a run-end-encoded array and reconstruct a table from it
        arrays = {
            'ree': pc.run_end_encode(result['ree']),
            'a': result['a'],
            'b': result['b'],
            'c': result['c'],
        }

        schema = pa.schema(
            [
                ("ree", arrays['ree'].type),
                ("a", arrays['a'].type),
                ("b", arrays['b'].type),
                ("c", arrays['c'].type),
            ]
        )
        arrow_tbl = pa.Table.from_arrays([arrays['ree'], arrays['a'], arrays['b'], arrays['c']], schema=schema)

        # Verify that the array is run end encoded
        ar_type = arrow_tbl['ree'].type
        assert pa.types.is_run_end_encoded(ar_type) == True

        # Scan the arrow table, making projections that don't cover the entire table
        # This should be pushed down into arrow to only provide us with the necessary columns

        res = duckdb_cursor.query(
            """
            select {} from arrow_tbl
        """.format(
                projection
            )
        ).arrow()

        # Verify correctness by fetching from the original table and the constructed result
        expected = duckdb_cursor.query("select {} from tbl".format(projection)).fetchall()
        actual = duckdb_cursor.query("select {} from res".format(projection)).fetchall()
        assert expected == actual

    def test_arrow_ree_list(self, duckdb_cursor):
        size = 1000
        duckdb_cursor.query(
            """
            create table tbl
            as select
                i // 4 as ree,
            FROM range({}) t(i)
        """.format(
                size
            )
        )

        # Populate the table with data
        unstructured = duckdb_cursor.query(
            """
            select * from tbl
        """
        ).arrow()

        columns = unstructured.columns
        # Run-encode the first column ('ree')
        columns[0] = pc.run_end_encode(columns[0])

        # Create a (chunked) ListArray from the chunked arrays (columns) of the ArrowTable
        structured_chunks = []
        for chunk in columns[0].iterchunks():
            ree = chunk
            chunk_length = len(ree)
            offsets = []
            offset = 0
            while chunk_length > 10:
                offsets.append(offset)
                offset += 10
                chunk_length -= 10

            new_array = pa.ListArray.from_arrays(offsets, values=ree)
            structured_chunks.append(new_array)

        structured = pa.chunked_array(structured_chunks)

        arrow_tbl = pa.Table.from_arrays([structured], names=['ree'])
        result = duckdb_cursor.query("select * from arrow_tbl").arrow()
        assert arrow_tbl.to_pylist() == result.to_pylist()

    def test_arrow_ree_struct(self, duckdb_cursor):
        duckdb_cursor.query(
            """
            create table tbl
            as select
                i // 4 as ree,
                i as a,
                i % 2 == 0 as b,
                i::VARCHAR as c
            FROM range(1000) t(i)
        """
        )

        # Populate the table with data
        unstructured = duckdb_cursor.query(
            """
            select * from tbl
        """
        ).arrow()

        columns = unstructured.columns
        # Run-encode the first column ('ree')
        columns[0] = pc.run_end_encode(columns[0])

        # Create a (chunked) StructArray from the chunked arrays (columns) of the ArrowTable
        names = unstructured.column_names
        iterables = [x.iterchunks() for x in columns]
        zipped = zip(*iterables)

        structured_chunks = [pa.StructArray.from_arrays([y for y in x], names=names) for x in zipped]
        structured = pa.chunked_array(structured_chunks)

        arrow_tbl = pa.Table.from_arrays([structured], names=['ree'])
        result = duckdb_cursor.query("select * from arrow_tbl").arrow()

        expected = duckdb_cursor.query("select {'ree': ree, 'a': a, 'b': b, 'c': c} as s from tbl").fetchall()
        actual = duckdb_cursor.query("select * from result").fetchall()

        assert expected == actual

    def test_arrow_ree_union(self, duckdb_cursor):
        size = 1000

        duckdb_cursor.query(
            """
            create table tbl
            as select
                i // 4 as ree,
                i as a,
                i % 2 == 0 as b,
                i::VARCHAR as c
            FROM range({}) t(i)
        """.format(
                size
            )
        )

        # Populate the table with data
        unstructured = duckdb_cursor.query(
            """
            select * from tbl
        """
        ).arrow()

        columns = unstructured.columns
        # Run-encode the first column ('ree')
        columns[0] = pc.run_end_encode(columns[0])

        # Create a (chunked) UnionArray from the chunked arrays (columns) of the ArrowTable
        names = unstructured.column_names
        iterables = [x.iterchunks() for x in columns]
        zipped = zip(*iterables)

        structured_chunks = []
        for chunk in zipped:
            ree, a, b, c = chunk
            chunk_length = len(ree)
            new_array = pa.UnionArray.from_sparse(
                pa.array([i % len(names) for i in range(chunk_length)], type=pa.int8()), [ree, a, b, c]
            )
            structured_chunks.append(new_array)

        structured = pa.chunked_array(structured_chunks)
        arrow_tbl = pa.Table.from_arrays([structured], names=['ree'])
        result = duckdb_cursor.query("select * from arrow_tbl").arrow()

        # Recreate the same result set
        expected = []
        for i in range(size):
            if i % 4 == 0:
                expected.append((i // 4,))
            elif i % 4 == 1:
                expected.append((i,))
            elif i % 4 == 2:
                expected.append((i % 2 == 0,))
            elif i % 4 == 3:
                expected.append((str(i),))
        actual = duckdb_cursor.query("select * from result").fetchall()
        assert expected == actual

    def test_arrow_ree_map(self, duckdb_cursor):
        size = 1000

        duckdb_cursor.query(
            """
            create table tbl
            as select
                i // 4 as ree,
                i as a,
            FROM range({}) t(i)
        """.format(
                size
            )
        )

        # Populate the table with data
        unstructured = duckdb_cursor.query(
            """
            select * from tbl
        """
        ).arrow()

        columns = unstructured.columns
        # Run-encode the first column ('ree')
        columns[0] = pc.run_end_encode(columns[0])

        # Create a (chunked) MapArray from the chunked arrays (columns) of the ArrowTable
        names = unstructured.column_names
        iterables = [x.iterchunks() for x in columns]
        zipped = zip(*iterables)

        structured_chunks = []
        for chunk in zipped:
            ree, a = chunk
            chunk_length = len(ree)
            offsets = []
            offset = 0
            while chunk_length > 10:
                offsets.append(offset)
                offset += 10
                chunk_length -= 10

            new_array = pa.MapArray.from_arrays(offsets, a, ree)
            structured_chunks.append(new_array)

        structured = pa.chunked_array(structured_chunks)
        arrow_tbl = pa.Table.from_arrays([structured], names=['ree'])
        result = duckdb_cursor.query("select * from arrow_tbl").arrow()

        # Verify that the resulting scan is the same as the input
        assert result.to_pylist() == arrow_tbl.to_pylist()

    def test_arrow_ree_dictionary(self, duckdb_cursor):
        size = 1000

        duckdb_cursor.query(
            """
            create table tbl
            as select
                i // 4 as ree,
            FROM range({}) t(i)
        """.format(
                size
            )
        )

        # Populate the table with data
        unstructured = duckdb_cursor.query(
            """
            select * from tbl
        """
        ).arrow()

        columns = unstructured.columns
        # Run-encode the first column ('ree')
        columns[0] = pc.run_end_encode(columns[0])

        # Create a (chunked) MapArray from the chunked arrays (columns) of the ArrowTable
        structured_chunks = []
        for chunk in columns[0].iterchunks():
            ree = chunk
            chunk_length = len(ree)
            offsets = [i for i in reversed(range(chunk_length))]

            new_array = pa.DictionaryArray.from_arrays(indices=offsets, dictionary=ree)
            structured_chunks.append(new_array)

        structured = pa.chunked_array(structured_chunks)
        arrow_tbl = pa.Table.from_arrays([structured], names=['ree'])
        result = duckdb_cursor.query("select * from arrow_tbl").arrow()

        # Verify that the resulting scan is the same as the input
        assert result.to_pylist() == arrow_tbl.to_pylist()



import duckdb
import numpy as np
import pytest
from conftest import NumpyPandas, ArrowPandas


class TestPandasMergeSameName(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_2304(self, duckdb_cursor, pandas):
        df1 = pandas.DataFrame(
            {
                'id_1': [1, 1, 1, 2, 2],
                'agedate': np.array(['2010-01-01', '2010-02-01', '2010-03-01', '2020-02-01', '2020-03-01']).astype(
                    'datetime64[D]'
                ),
                'age': [1, 2, 3, 1, 2],
                'v': [1.1, 1.2, 1.3, 2.1, 2.2],
            }
        )

        df2 = pandas.DataFrame(
            {
                'id_1': [1, 1, 2],
                'agedate': np.array(['2010-01-01', '2010-02-01', '2020-03-01']).astype('datetime64[D]'),
                'v2': [11.1, 11.2, 21.2],
            }
        )

        con = duckdb.connect()
        con.register('df1', df1)
        con.register('df2', df2)
        query = """SELECT * from df1
        LEFT  OUTER JOIN df2
        ON (df1.id_1=df2.id_1 and df1.agedate=df2.agedate)  order by df1.id_1, df1.agedate, df1.age, df1.v, df2.id_1,df2.agedate,df2.v2"""

        result_df = con.execute(query).fetchdf()
        expected_result = con.execute(query).fetchall()
        con.register('result_df', result_df)
        rel = con.sql(
            """
            select * from result_df order by
                id_1,
                agedate,
                age,
                v,
                id_1_1,
                agedate_1,
                v2
        """
        )
        result = rel.fetchall()

        assert result == expected_result

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_pd_names(self, duckdb_cursor, pandas):
        df1 = pandas.DataFrame(
            {
                'id': [1, 1, 2],
                'id_1': [1, 1, 2],
                'id_3': [1, 1, 2],
            }
        )

        df2 = pandas.DataFrame({'id': [1, 1, 2], 'id_1': [1, 1, 2], 'id_2': [1, 1, 1]})

        exp_result = pandas.DataFrame(
            {
                'id': [1, 1, 2, 1, 1],
                'id_1': [1, 1, 2, 1, 1],
                'id_3': [1, 1, 2, 1, 1],
                'id_2': [1, 1, 2, 1, 1],
                'id_1_1': [1, 1, 2, 1, 1],
                'id_2_1': [1, 1, 1, 1, 1],
            }
        )

        con = duckdb.connect()
        con.register('df1', df1)
        con.register('df2', df2)
        query = """SELECT * from df1
        LEFT OUTER JOIN df2
        ON (df1.id_1=df2.id_1)"""

        result_df = con.execute(query).fetchdf()
        pandas.testing.assert_frame_equal(exp_result, result_df)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_repeat_name(self, duckdb_cursor, pandas):
        df1 = pandas.DataFrame(
            {
                'id': [1],
                'id_1': [1],
                'id_2': [1],
            }
        )

        df2 = pandas.DataFrame({'id': [1]})

        exp_result = pandas.DataFrame(
            {
                'id': [1],
                'id_1': [1],
                'id_2': [1],
                'id_3': [1],
            }
        )

        con = duckdb.connect()
        con.register('df1', df1)
        con.register('df2', df2)

        result_df = con.execute(
            """
                SELECT * from df1
                LEFT OUTER JOIN df2
                ON (df1.id=df2.id)
            """
        ).fetchdf()
        pandas.testing.assert_frame_equal(exp_result, result_df)



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import (
    LongType,
    StructType,
    BooleanType,
    StructField,
    StringType,
    IntegerType,
    LongType,
    Row,
    ArrayType,
    MapType,
)
from duckdb.experimental.spark.sql.functions import col, struct, when, lit
import duckdb
import re


class TestWithColumn(object):
    def test_with_column(self, spark):
        data = [
            ('James', '', 'Smith', '1991-04-01', 'M', 3000),
            ('Michael', 'Rose', '', '2000-05-19', 'M', 4000),
            ('Robert', '', 'Williams', '1978-09-05', 'M', 4000),
            ('Maria', 'Anne', 'Jones', '1967-12-01', 'F', 4000),
            ('Jen', 'Mary', 'Brown', '1980-02-17', 'F', -1),
        ]

        columns = ["firstname", "middlename", "lastname", "dob", "gender", "salary"]
        df = spark.createDataFrame(data=data, schema=columns)
        assert df.schema['salary'].dataType.typeName() == 'integer'

        # The type of 'salary' has been cast to Bigint
        new_df = df.withColumn("salary", col("salary").cast("BIGINT"))
        assert new_df.schema['salary'].dataType.typeName() == 'long'

        # Replace the 'salary' column with '(salary * 100)'
        df2 = df.withColumn("salary", col("salary") * 100)
        res = df2.collect()
        assert res[0].salary == 300_000

        # Create a column from an existing column
        df2 = df.withColumn("CopiedColumn", col("salary") * -1)
        res = df2.collect()
        assert res[0].CopiedColumn == -3000

        df2 = df.withColumn("Country", lit("USA"))
        res = df2.collect()
        assert res[0].Country == 'USA'

        df2 = df.withColumn("Country", lit("USA")).withColumn("anotherColumn", lit("anotherValue"))
        res = df2.collect()
        assert res[0].Country == 'USA'
        assert res[1].anotherColumn == 'anotherValue'

        df2 = df.withColumnRenamed("gender", "sex")
        assert 'gender' not in df2
        assert 'sex' in df2

        df2 = df.drop("salary")
        assert 'salary' not in df2



from duckdb.duckdb.functional import (
	FunctionNullHandling,
	PythonUDFType,
	SPECIAL,
	DEFAULT,
	NATIVE,
	ARROW
)

__all__ = [
	"FunctionNullHandling",
	"PythonUDFType",
	"SPECIAL",
	"DEFAULT",
	"NATIVE",
	"ARROW"
]



import duckdb
from threading import Thread, current_thread
import pandas as pd
import os
import pytest


@pytest.fixture(scope="session")
def tmp_database(tmp_path_factory):
    database = tmp_path_factory.mktemp("databases", numbered=True) / "tmp.duckdb"
    return str(database)


def insert_from_cursor(duckdb_con):
    # Insert a row with the name of the thread
    duckdb_cursor = duckdb_con.cursor()  # Make a cursor within the thread
    thread_name = str(current_thread().name)
    duckdb_cursor.execute("""INSERT INTO my_inserts VALUES (?)""", (thread_name,))


def insert_from_same_connection(duckdb_cursor):
    # Insert a row with the name of the thread
    thread_name = str(current_thread().name)
    duckdb_cursor.execute("""INSERT INTO my_inserts VALUES (?)""", (thread_name,))


class TestPythonMultithreading(object):
    def test_multiple_cursors(self, duckdb_cursor):
        duckdb_con = duckdb.connect()  # In Memory DuckDB
        duckdb_con.execute("""CREATE OR REPLACE TABLE my_inserts (thread_name varchar)""")

        thread_count = 3
        threads = []

        # Kick off multiple threads (in the same process)
        # Pass in the same connection as an argument, and an object to store the results
        for i in range(thread_count):
            threads.append(Thread(target=insert_from_cursor, args=(duckdb_con,), name='my_thread_' + str(i)))

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        assert duckdb_con.execute("""SELECT * FROM my_inserts order by thread_name""").fetchall() == [
            ('my_thread_0',),
            ('my_thread_1',),
            ('my_thread_2',),
        ]

    def test_same_connection(self, duckdb_cursor):
        duckdb_con = duckdb.connect()  # In Memory DuckDB
        duckdb_con.execute("""CREATE OR REPLACE TABLE my_inserts (thread_name varchar)""")

        thread_count = 3
        threads = []
        cursors = []

        # Kick off multiple threads (in the same process)
        # Pass in the same connection as an argument, and an object to store the results
        for i in range(thread_count):
            cursors.append(duckdb_con.cursor())
            threads.append(Thread(target=insert_from_same_connection, args=(cursors[i],), name='my_thread_' + str(i)))

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        assert duckdb_con.execute("""SELECT * FROM my_inserts order by thread_name""").fetchall() == [
            ('my_thread_0',),
            ('my_thread_1',),
            ('my_thread_2',),
        ]

    def test_multiple_cursors_persisted(self, tmp_database):
        duckdb_con = duckdb.connect(tmp_database)
        duckdb_con.execute("""CREATE OR REPLACE TABLE my_inserts (thread_name varchar)""")

        thread_count = 3
        threads = []

        # Kick off multiple threads (in the same process)
        # Pass in the same connection as an argument, and an object to store the results
        for i in range(thread_count):
            threads.append(Thread(target=insert_from_cursor, args=(duckdb_con,), name='my_thread_' + str(i)))
        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        assert duckdb_con.execute("""SELECT * FROM my_inserts order by thread_name""").fetchall() == [
            ('my_thread_0',),
            ('my_thread_1',),
            ('my_thread_2',),
        ]
        duckdb_con.close()

    def test_same_connection_persisted(self, tmp_database):
        duckdb_con = duckdb.connect(tmp_database)
        duckdb_con.execute("""CREATE OR REPLACE TABLE my_inserts (thread_name varchar)""")

        thread_count = 3
        threads = []

        # Kick off multiple threads (in the same process)
        # Pass in the same connection as an argument, and an object to store the results
        for i in range(thread_count):
            threads.append(Thread(target=insert_from_same_connection, args=(duckdb_con,), name='my_thread_' + str(i)))
        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        assert duckdb_con.execute("""SELECT * FROM my_inserts order by thread_name""").fetchall() == [
            ('my_thread_0',),
            ('my_thread_1',),
            ('my_thread_2',),
        ]
        duckdb_con.close()



import pytest
import tempfile

import os

_ = pytest.importorskip("duckdb.experimental.spark")


@pytest.fixture
def df(spark):
    simpleData = (
        ("Java", 4000, 5),
        ("Python", 4600, 10),
        ("Scala", 4100, 15),
        ("Scala", 4500, 15),
        ("PHP", 3000, 20),
    )
    columns = ["CourseName", "fee", "discount"]
    dataframe = spark.createDataFrame(data=simpleData, schema=columns)
    yield dataframe


class TestSparkToParquet(object):
    def test_basic_to_parquet(self, df, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.parquet")

        df.write.parquet(temp_file_name)

        csv_rel = spark.read.parquet(temp_file_name)

        assert df.collect() == csv_rel.collect()

    def test_compressed_to_parquet(self, df, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.parquet")

        df.write.parquet(temp_file_name, compression="ZSTD")

        csv_rel = spark.read.parquet(temp_file_name)

        assert df.collect() == csv_rel.collect()



import pytest
import tempfile
import gc
import duckdb
import os

try:
    import pyarrow
    import pyarrow.parquet

    can_run = True
except:
    can_run = False


class TestArrowUnregister(object):
    def test_arrow_unregister1(self, duckdb_cursor):
        if not can_run:
            return
        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')
        cols = 'id, first_name, last_name, email, gender, ip_address, cc, country, birthdate, salary, title, comments'

        arrow_table_obj = pyarrow.parquet.read_table(parquet_filename)
        connection = duckdb.connect(":memory:")
        connection.register("arrow_table", arrow_table_obj)

        arrow_table_2 = connection.execute("SELECT * FROM arrow_table;").fetch_arrow_table()
        connection.unregister("arrow_table")
        with pytest.raises(duckdb.CatalogException, match='Table with name arrow_table does not exist'):
            connection.execute("SELECT * FROM arrow_table;").fetch_arrow_table()
        with pytest.raises(duckdb.CatalogException, match='View with name arrow_table does not exist'):
            connection.execute("DROP VIEW arrow_table;")
        connection.execute("DROP VIEW IF EXISTS arrow_table;")

    def test_arrow_unregister2(self, duckdb_cursor):
        if not can_run:
            return
        fd, db = tempfile.mkstemp()
        os.close(fd)
        os.remove(db)

        connection = duckdb.connect(db)
        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')
        cols = 'id, first_name, last_name, email, gender, ip_address, cc, country, birthdate, salary, title, comments'
        arrow_table_obj = pyarrow.parquet.read_table(parquet_filename)
        connection.register("arrow_table", arrow_table_obj)
        connection.unregister("arrow_table")  # Attempting to unregister.
        connection.close()
        # Reconnecting while Arrow Table still in mem.
        connection = duckdb.connect(db)
        assert len(connection.execute("PRAGMA show_tables;").fetchall()) == 0
        with pytest.raises(duckdb.CatalogException, match='Table with name arrow_table does not exist'):
            connection.execute("SELECT * FROM arrow_table;").fetch_arrow_table()
        connection.close()
        del arrow_table_obj
        gc.collect()
        # Reconnecting after Arrow Table is freed.
        connection = duckdb.connect(db)
        assert len(connection.execute("PRAGMA show_tables;").fetchall()) == 0
        with pytest.raises(duckdb.CatalogException, match='Table with name arrow_table does not exist'):
            connection.execute("SELECT * FROM arrow_table;").fetch_arrow_table()
        connection.close()



import duckdb
import pytest
from decimal import Decimal

pa = pytest.importorskip("pyarrow")


class TestArrowDecimal256(object):
    def test_decimal_256_throws(self, duckdb_cursor):
        with duckdb.connect() as conn:
            pa_decimal256 = pa.Table.from_pylist(
                [{"data": Decimal("100.00")} for _ in range(4)],
                pa.schema([("data", pa.decimal256(12, 4))]),
            )
            with pytest.raises(
                duckdb.NotImplementedException, match="Unsupported Internal Arrow Type for Decimal d:12,4,256"
            ):
                conn.execute("select * from pa_decimal256;").arrow().to_pylist()



import pytest
import duckdb

VECTOR_SIZE = duckdb.__standard_vector_size__


class TestType(object):
    def test_fetch_df_chunk(self):
        size = 3000
        con = duckdb.connect()
        con.execute(f"CREATE table t as select range a from range({size});")
        query = con.execute("SELECT a FROM t")

        # Fetch the first chunk
        cur_chunk = query.fetch_df_chunk()
        assert cur_chunk['a'][0] == 0
        assert len(cur_chunk) == VECTOR_SIZE

        # Fetch the second chunk, can't be entirely filled
        cur_chunk = query.fetch_df_chunk()
        assert cur_chunk['a'][0] == VECTOR_SIZE
        expected = size - VECTOR_SIZE
        assert len(cur_chunk) == expected

    @pytest.mark.parametrize('size', [3000, 10000, 100000, VECTOR_SIZE - 1, VECTOR_SIZE + 1, VECTOR_SIZE])
    def test_monahan(self, size):
        con = duckdb.connect()
        con.execute(f"CREATE table t as select range a from range({size});")
        query = con.execute("SELECT a FROM t")

        processed = 0
        expected_chunks = size // VECTOR_SIZE
        if size % VECTOR_SIZE != 0:
            expected_chunks += 1

        for _ in range(expected_chunks):
            expected_size = size - processed
            if expected_size > VECTOR_SIZE:
                expected_size = VECTOR_SIZE
            cur_chunk = query.fetch_df_chunk()
            assert len(cur_chunk) == expected_size
            processed += expected_size

        cur_chunk = query.fetch_df_chunk()
        assert len(cur_chunk) == 0

    def test_fetch_df_chunk_parameter(self):
        size = 10000
        con = duckdb.connect()
        con.execute(f"CREATE table t as select range a from range({size});")
        query = con.execute("SELECT a FROM t")

        # Return 2 vectors
        cur_chunk = query.fetch_df_chunk(2)
        assert cur_chunk['a'][0] == 0
        assert len(cur_chunk) == VECTOR_SIZE * 2

        # Return Default 1 vector
        cur_chunk = query.fetch_df_chunk()
        assert cur_chunk['a'][0] == VECTOR_SIZE * 2
        assert len(cur_chunk) == VECTOR_SIZE

        # Return 0 vectors
        cur_chunk = query.fetch_df_chunk(0)
        assert len(cur_chunk) == 0

        fetched = VECTOR_SIZE * 3
        expected = size - fetched

        # Return more vectors than we have remaining
        cur_chunk = query.fetch_df_chunk(3)
        assert cur_chunk['a'][0] == fetched
        assert len(cur_chunk) == expected

        # These shouldn't throw errors (Just emmit empty chunks)
        cur_chunk = query.fetch_df_chunk(100)
        assert len(cur_chunk) == 0

        cur_chunk = query.fetch_df_chunk(0)
        assert len(cur_chunk) == 0

        cur_chunk = query.fetch_df_chunk()
        assert len(cur_chunk) == 0

    def test_fetch_df_chunk_negative_parameter(self):
        con = duckdb.connect()
        con.execute("CREATE table t as select range a from range(100);")
        query = con.execute("SELECT a FROM t")

        # Return -1 vector should not work
        with pytest.raises(TypeError, match='incompatible function arguments'):
            cur_chunk = query.fetch_df_chunk(-1)



from .sql import SparkSession, DataFrame
from .conf import SparkConf
from .context import SparkContext
from ._globals import _NoValue
from .exception import ContributionsAcceptedError

__all__ = ["SparkSession", "DataFrame", "SparkConf", "SparkContext", "ContributionsAcceptedError"]



import traceback


class TestNull(object):
    def test_fetchone_null(self, duckdb_cursor):
        duckdb_cursor.execute("CREATE TABLE atable (Value int)")
        duckdb_cursor.execute("INSERT INTO atable VALUES (1)")
        duckdb_cursor.execute("SELECT * FROM atable")
        assert duckdb_cursor.fetchone()[0] == 1
        assert duckdb_cursor.fetchone() is None



import duckdb
import pytest
from os import path
import shutil
import os
from pathlib import Path


def export_database(export_location):
    # Create the db
    con = duckdb.connect()
    con.execute("create table tbl (a integer, b integer);")
    con.execute("insert into tbl values (5,1);")

    # Export the db
    con.execute(f"export database '{export_location}';")
    print(f"Exported database to {export_location}")


def import_database(import_location):
    con = duckdb.connect()
    con.execute(f"import database '{import_location}'")
    print(f"Imported database from {import_location}")

    res = con.query("select * from tbl").fetchall()
    assert res == [
        (5, 1),
    ]
    print("Successfully queried an imported database that was moved from its original export location!")


def move_database(export_location, import_location):
    assert path.exists(export_location)
    assert path.exists(import_location)

    for file in ['schema.sql', 'load.sql', 'tbl.csv']:
        shutil.move(path.join(export_location, file), import_location)


def export_move_and_import(export_path, import_path):
    export_database(export_path)
    move_database(export_path, import_path)
    import_database(import_path)


def export_and_import_empty_db(db_path, _):
    con = duckdb.connect()

    # Export the db
    con.execute(f"export database '{db_path}';")
    print(f"Exported database to {db_path}")

    con.close()
    con = duckdb.connect()
    con.execute(f"import database '{db_path}'")


class TestDuckDBImportExport:
    @pytest.mark.parametrize('routine', [export_move_and_import, export_and_import_empty_db])
    def test_import_and_export(self, routine, tmp_path_factory):
        export_path = str(tmp_path_factory.mktemp("export_dbs", numbered=True))
        import_path = str(tmp_path_factory.mktemp("import_dbs", numbered=True))
        routine(export_path, import_path)

    def test_import_empty_db(self, tmp_path_factory):
        import_path = str(tmp_path_factory.mktemp("empty_db", numbered=True))

        # Create an empty db folder structure
        Path(Path(import_path) / 'load.sql').touch()
        Path(Path(import_path) / 'schema.sql').touch()

        con = duckdb.connect()
        con.execute(f"import database '{import_path}'")

        # Put a single comment into the 'schema.sql' file
        with open(Path(import_path) / 'schema.sql', 'w') as f:
            f.write('--\n')

        con.close()
        con = duckdb.connect()
        con.execute(f"import database '{import_path}'")



import pytest

pd = pytest.importorskip("pandas")
np = pytest.importorskip("numpy")

NUMBER_OF_ROWS = 200000
NUMBER_OF_COLUMNS = 1


class TestMetaTransaction(object):
    def test_fetchmany(self, duckdb_cursor):
        duckdb_cursor.execute("CREATE SEQUENCE id_seq")
        column_names = ',\n'.join([f'column_{i} FLOAT' for i in range(1, NUMBER_OF_COLUMNS + 1)])
        create_table_query = f"""
        CREATE TABLE my_table (
            id INTEGER DEFAULT nextval('id_seq'),
            {column_names}
        )
        """
        # Create a table containing a sequence
        duckdb_cursor.execute(create_table_query)

        for i in range(20):
            # Then insert a large amount of tuples, triggering a parallel execution
            data = np.random.rand(NUMBER_OF_ROWS, NUMBER_OF_COLUMNS)
            columns = [f'Column_{i+1}' for i in range(NUMBER_OF_COLUMNS)]
            df = pd.DataFrame(data, columns=columns)
            df_columns = ", ".join(df.columns)
            # This gets executed in parallel, causing NextValFunction to be called in parallel
            # stressing the MetaTransaction::Get concurrency
            duckdb_cursor.execute(f"INSERT INTO my_table ({df_columns}) SELECT * FROM df")
            print(f"inserted {i}")
            duckdb_cursor.commit()



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")
from duckdb.experimental.spark.sql.types import Row
from duckdb.experimental.spark.sql import functions as F


class TestNumericFunctions(object):
    def test_greatest(self, spark):
        data = [
            (1, 2),
            (4, 3),
        ]
        df = spark.createDataFrame(data, ["firstColumn", "secondColumn"])
        df = df.withColumn("greatest_value", F.greatest(F.col("firstColumn"), F.col("secondColumn")))
        res = df.select("greatest_value").collect()
        assert res == [
            Row(greatest_value=2),
            Row(greatest_value=4),
        ]

    def test_least(self, spark):
        data = [
            (1, 2),
            (4, 3),
        ]
        df = spark.createDataFrame(data, ["firstColumn", "secondColumn"])
        df = df.withColumn("least_value", F.least(F.col("firstColumn"), F.col("secondColumn")))
        res = df.select("least_value").collect()
        assert res == [
            Row(least_value=1),
            Row(least_value=3),
        ]

    def test_ceil(self, spark):
        data = [
            (1.1,),
            (2.9,),
        ]
        df = spark.createDataFrame(data, ["firstColumn"])
        df = df.withColumn("ceil_value", F.ceil(F.col("firstColumn")))
        res = df.select("ceil_value").collect()
        assert res == [
            Row(ceil_value=2),
            Row(ceil_value=3),
        ]

    def test_floor(self, spark):
        data = [
            (1.1,),
            (2.9,),
        ]
        df = spark.createDataFrame(data, ["firstColumn"])
        df = df.withColumn("floor_value", F.floor(F.col("firstColumn")))
        res = df.select("floor_value").collect()
        assert res == [
            Row(floor_value=1),
            Row(floor_value=2),
        ]

    def test_abs(self, spark):
        data = [
            (1.1,),
            (-2.9,),
        ]
        df = spark.createDataFrame(data, ["firstColumn"])
        df = df.withColumn("abs_value", F.abs(F.col("firstColumn")))
        res = df.select("abs_value").collect()
        assert res == [
            Row(abs_value=1.1),
            Row(abs_value=2.9),
        ]

    def test_sqrt(self, spark):
        data = [
            (4,),
            (9,),
        ]
        df = spark.createDataFrame(data, ["firstColumn"])
        df = df.withColumn("sqrt_value", F.sqrt(F.col("firstColumn")))
        res = df.select("sqrt_value").collect()
        assert res == [
            Row(sqrt_value=2.0),
            Row(sqrt_value=3.0),
        ]



import duckdb
import pytest


class TestRAPIDescription(object):
    def test_rapi_description(self, duckdb_cursor):
        res = duckdb_cursor.query('select 42::INT AS a, 84::BIGINT AS b')
        desc = res.description
        names = [x[0] for x in desc]
        types = [x[1] for x in desc]
        assert names == ['a', 'b']
        assert types == ['NUMBER', 'NUMBER']

    def test_rapi_describe(self, duckdb_cursor):
        np = pytest.importorskip("numpy")
        pd = pytest.importorskip("pandas")
        res = duckdb_cursor.query('select 42::INT AS a, 84::BIGINT AS b')
        duck_describe = res.describe().df()
        np.testing.assert_array_equal(duck_describe['aggr'], ['count', 'mean', 'stddev', 'min', 'max', 'median'])
        np.testing.assert_array_equal(duck_describe['a'], [1, 42, float('nan'), 42, 42, 42])
        np.testing.assert_array_equal(duck_describe['b'], [1, 84, float('nan'), 84, 84, 84])

        # now with more values
        res = duckdb_cursor.query(
            'select CASE WHEN i%2=0 THEN i ELSE NULL END AS i, i * 10 AS j, (i * 23 // 27)::DOUBLE AS k FROM range(10000) t(i)'
        )
        duck_describe = res.describe().df()
        np.testing.assert_allclose(duck_describe['i'], [5000.0, 4999.0, 2887.0400066504103, 0.0, 9998.0, 4999.0])
        np.testing.assert_allclose(duck_describe['j'], [10000.0, 49995.0, 28868.956799071675, 0.0, 99990.0, 49995.0])
        np.testing.assert_allclose(duck_describe['k'], [10000.0, 4258.3518, 2459.207430770227, 0.0, 8517.0, 4258.5])

        # describe data with other (non-numeric) types
        res = duckdb_cursor.query("select 'hello world' AS a, [1, 2, 3] AS b")
        duck_describe = res.describe().df()
        assert len(duck_describe) > 0

        # describe mixed table
        res = duckdb_cursor.query("select 42::INT AS a, 84::BIGINT AS b, 'hello world' AS c")
        duck_describe = res.describe().df()
        np.testing.assert_array_equal(duck_describe['a'], [1, 42, float('nan'), 42, 42, 42])
        np.testing.assert_array_equal(duck_describe['b'], [1, 84, float('nan'), 84, 84, 84])

        # timestamps
        res = duckdb_cursor.query("select timestamp '1992-01-01', date '2000-01-01'")
        duck_describe = res.describe().df()
        assert len(duck_describe) > 0

        # describe empty result
        res = duckdb_cursor.query("select 42 AS a LIMIT 0")
        duck_describe = res.describe().df()
        assert len(duck_describe) > 0



import duckdb

try:
    import pyarrow as pa
    import pyarrow.parquet
    import numpy as np
    import pandas as pd
    import pytest

    can_run = True
except:
    can_run = False


def compare_results(duckdb_cursor, query):
    true_answer = duckdb_cursor.query(query).fetchall()
    from_arrow = duckdb.from_arrow(duckdb_cursor.query(query).arrow()).fetchall()
    assert true_answer == from_arrow


def arrow_to_pandas(duckdb_cursor, query):
    return duckdb_cursor.query(query).arrow().to_pandas()['a'].values.tolist()


class TestArrowNested(object):
    def test_lists_basic(self, duckdb_cursor):
        if not can_run:
            return

        # Test Constant List
        query = duckdb_cursor.query("SELECT a from (select list_value(3,5,10) as a) as t").arrow()['a'].to_numpy()
        assert query[0][0] == 3
        assert query[0][1] == 5
        assert query[0][2] == 10

        # Empty List
        query = duckdb_cursor.query("SELECT a from (select list_value() as a) as t").arrow()['a'].to_numpy()
        assert len(query[0]) == 0

        # Test Constant List With Null
        query = duckdb_cursor.query("SELECT a from (select list_value(3,NULL) as a) as t").arrow()['a'].to_numpy()
        assert query[0][0] == 3
        assert np.isnan(query[0][1])

    def test_list_types(self, duckdb_cursor):
        if not can_run:
            return

        # Large Lists
        data = pyarrow.array([[1], None, [2]], type=pyarrow.large_list(pyarrow.int64()))
        arrow_table = pa.Table.from_arrays([data], ['a'])
        rel = duckdb.from_arrow(arrow_table)
        res = rel.execute().fetchall()
        assert res == [([1],), (None,), ([2],)]

        # Fixed Size Lists
        data = pyarrow.array([[1], None, [2]], type=pyarrow.list_(pyarrow.int64(), 1))
        arrow_table = pa.Table.from_arrays([data], ['a'])
        rel = duckdb.from_arrow(arrow_table)
        res = rel.execute().fetchall()
        assert res == [((1,),), (None,), ((2,),)]

        # Complex nested structures with different list types
        data = [
            pyarrow.array([[1], None, [2]], type=pyarrow.list_(pyarrow.int64(), 1)),
            pyarrow.array([[1], None, [2]], type=pyarrow.large_list(pyarrow.int64())),
            pyarrow.array([[1, 2, 3], None, [2, 1]], type=pyarrow.list_(pyarrow.int64())),
        ]
        arrow_table = pa.Table.from_arrays([data[0], data[1], data[2]], ['a', 'b', 'c'])
        rel = duckdb.from_arrow(arrow_table)
        res = rel.project('a').execute().fetchall()
        assert res == [((1,),), (None,), ((2,),)]
        res = rel.project('b').execute().fetchall()
        assert res == [([1],), (None,), ([2],)]
        res = rel.project('c').execute().fetchall()
        assert res == [([1, 2, 3],), (None,), ([2, 1],)]

        # Struct Holding different List Types
        struct = [pa.StructArray.from_arrays(data, ['fixed', 'large', 'normal'])]
        arrow_table = pa.Table.from_arrays(struct, ['a'])
        rel = duckdb.from_arrow(arrow_table)
        res = rel.execute().fetchall()
        assert res == [
            ({'fixed': (1,), 'large': [1], 'normal': [1, 2, 3]},),
            ({'fixed': None, 'large': None, 'normal': None},),
            ({'fixed': (2,), 'large': [2], 'normal': [2, 1]},),
        ]

    def test_lists_roundtrip(self, duckdb_cursor):
        if not can_run:
            return
        # Integers
        compare_results(duckdb_cursor, "SELECT a from (select list_value(3,5,10) as a) as t")
        compare_results(duckdb_cursor, "SELECT a from (select list_value(3,5,NULL) as a) as t")
        compare_results(duckdb_cursor, "SELECT a from (select list_value(NULL,NULL,NULL) as a) as t")
        compare_results(duckdb_cursor, "SELECT a from (select list_value() as a) as t")
        # Strings
        compare_results(duckdb_cursor, "SELECT a from (select list_value('test','test_one','test_two') as a) as t")
        compare_results(duckdb_cursor, "SELECT a from (select list_value('test','test_one',NULL) as a) as t")
        # Big Lists
        compare_results(duckdb_cursor, "SELECT a from (SELECT LIST(i) as a FROM range(10000) tbl(i)) as t")
        # Multiple Lists
        compare_results(
            duckdb_cursor,
            "SELECT a from (SELECT LIST(i) as a FROM range(10000) tbl(i) group by i%10 order by all) as t",
        )
        # Unique Constants
        compare_results(duckdb_cursor, "SELECT a from (SELECT list_value(1) as a FROM range(10) tbl(i)) as t")
        # Nested Lists
        compare_results(
            duckdb_cursor,
            "SELECT LIST(le order by le) FROM (SELECT LIST(i order by i) le from range(100) tbl(i) group by i%10) as t",
        )

        # LIST[LIST[LIST[LIST[LIST[INTEGER]]]]]]
        compare_results(
            duckdb_cursor,
            "SELECT list (lllle order by lllle) llllle from (SELECT list (llle order by llle) lllle from (SELECT list(lle order by lle) llle from (SELECT LIST(le order by le) lle FROM (SELECT LIST(i order by i) le from range(100) tbl(i) group by i%10) as t) as t1) as t2) as t3",
        )

        compare_results(
            duckdb_cursor,
            '''SELECT grp,lst,cs FROM (select grp, lst, case when grp>1 then lst else list_value(null) end as cs
                        from (SELECT a%4 as grp, list(a order by a) as lst FROM range(7) tbl(a) group by grp) as lst_tbl) as T order by all;''',
        )
        # Tests for converting multiple lists to/from Arrow with NULL values and/or strings
        compare_results(
            duckdb_cursor,
            "SELECT list(st order by st) from (select i, case when i%10 then NULL else i::VARCHAR end as st from range(1000) tbl(i)) as t group by i%5 order by all",
        )

    def test_struct_roundtrip(self, duckdb_cursor):
        if not can_run:
            return
        compare_results(duckdb_cursor, "SELECT a from (SELECT STRUCT_PACK(a := 42, b := 43) as a) as t")
        compare_results(duckdb_cursor, "SELECT a from (SELECT STRUCT_PACK(a := NULL, b := 43) as a) as t")
        compare_results(duckdb_cursor, "SELECT a from (SELECT STRUCT_PACK(a := NULL) as a) as t")
        compare_results(
            duckdb_cursor, "SELECT a from (SELECT STRUCT_PACK(a := i, b := i) as a FROM range(10000) tbl(i)) as t"
        )
        compare_results(
            duckdb_cursor,
            "SELECT a from (SELECT STRUCT_PACK(a := LIST_VALUE(1,2,3), b := i) as a FROM range(10000) tbl(i)) as t",
        )

    def test_map_roundtrip(self, duckdb_cursor):
        if not can_run:
            return
        compare_results(
            duckdb_cursor, "SELECT a from (select MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as a) as t"
        )

        compare_results(
            duckdb_cursor, "SELECT a from (select MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as a) as t"
        )

        compare_results(duckdb_cursor, "SELECT a from (select MAP(LIST_VALUE(),LIST_VALUE()) as a) as t")
        compare_results(
            duckdb_cursor,
            "SELECT a from (select MAP(LIST_VALUE('Jon Lajoie', 'Backstreet Boys', 'Tenacious D'),LIST_VALUE(10,9,10)) as a) as t",
        )
        compare_results(
            duckdb_cursor,
            "SELECT a from (select MAP(LIST_VALUE('Jon Lajoie','Tenacious D'),LIST_VALUE(10,10)) as a) as t",
        )
        compare_results(
            duckdb_cursor, "SELECT m from (select MAP(list_value(1), list_value(2)) from range(5) tbl(i)) tbl(m)"
        )
        compare_results(
            duckdb_cursor,
            "SELECT m from (select MAP(lsta,lstb) as m from (SELECT list(i) as lsta, list(i) as lstb from range(10000) tbl(i) group by i%5 order by all) as lst_tbl) as T",
        )

    def test_map_arrow_to_duckdb(self, duckdb_cursor):
        if not can_run:
            return
        map_type = pa.map_(pa.int32(), pa.int32())
        values = [[(3, 12), (3, 21)], [(5, 42)]]
        arrow_table = pa.table({'detail': pa.array(values, map_type)})
        with pytest.raises(
            duckdb.InvalidInputException,
            match="Arrow map contains duplicate key, which isn't supported by DuckDB map type",
        ):
            rel = duckdb.from_arrow(arrow_table).fetchall()

    def test_map_arrow_to_pandas(self, duckdb_cursor):
        if not can_run:
            return
        assert arrow_to_pandas(
            duckdb_cursor, "SELECT a from (select MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as a) as t"
        ) == [[(1, 10), (2, 9), (3, 8), (4, 7)]]
        assert arrow_to_pandas(duckdb_cursor, "SELECT a from (select MAP(LIST_VALUE(),LIST_VALUE()) as a) as t") == [[]]
        assert arrow_to_pandas(
            duckdb_cursor,
            "SELECT a from (select MAP(LIST_VALUE('Jon Lajoie', 'Backstreet Boys', 'Tenacious D'),LIST_VALUE(10,9,10)) as a) as t",
        ) == [[('Jon Lajoie', 10), ('Backstreet Boys', 9), ('Tenacious D', 10)]]
        assert arrow_to_pandas(
            duckdb_cursor, "SELECT a from (select MAP(list_value(1), list_value(2)) from range(5) tbl(i)) tbl(a)"
        ) == [[(1, 2)], [(1, 2)], [(1, 2)], [(1, 2)], [(1, 2)]]
        assert arrow_to_pandas(
            duckdb_cursor,
            "SELECT MAP(LIST_VALUE({'i':1,'j':2},{'i':3,'j':4}),LIST_VALUE({'i':1,'j':2},{'i':3,'j':4})) as a",
        ) == [[({'i': 1, 'j': 2}, {'i': 1, 'j': 2}), ({'i': 3, 'j': 4}, {'i': 3, 'j': 4})]]

    def test_frankstein_nested(self, duckdb_cursor):
        if not can_run:
            return
        # List of structs W/ Struct that is NULL entirely
        compare_results(duckdb_cursor, "SELECT [{'i':1,'j':2},NULL,{'i':2,'j':NULL}]")

        # Lists of structs with lists
        compare_results(duckdb_cursor, "SELECT [{'i':1,'j':[2,3]},NULL]")

        # Maps embedded in a struct
        compare_results(
            duckdb_cursor,
            "SELECT {'i':mp,'j':mp2} FROM (SELECT MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as mp, MAP(LIST_VALUE(1, 2, 3, 5),LIST_VALUE(10, 9, 8, 7)) as mp2) as t",
        )

        # List of maps
        compare_results(
            duckdb_cursor,
            "SELECT [mp,mp2] FROM (SELECT MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7)) as mp, MAP(LIST_VALUE(1, 2, 3, 5),LIST_VALUE(10, 9, 8, 7)) as mp2) as t",
        )

        # Map with list as key and/or value
        compare_results(duckdb_cursor, "SELECT MAP(LIST_VALUE([1,2],[3,4],[5,4]),LIST_VALUE([1,2],[3,4],[5,4]))")

        # Map with struct as key and/or value
        compare_results(
            duckdb_cursor, "SELECT MAP(LIST_VALUE({'i':1,'j':2},{'i':3,'j':4}),LIST_VALUE({'i':1,'j':2},{'i':3,'j':4}))"
        )

        # Struct that is NULL entirely
        compare_results(duckdb_cursor, "SELECT * FROM (VALUES ({'i':1,'j':2}), (NULL), ({'i':1,'j':2}), (NULL)) as a")

        # Null checks on lists with structs
        compare_results(duckdb_cursor, "SELECT [{'i':1,'j':[2,3]},NULL,{'i':1,'j':[2,3]}]")

        # MAP that is NULL entirely
        compare_results(
            duckdb_cursor,
            "SELECT * FROM (VALUES (MAP(LIST_VALUE(1,2),LIST_VALUE(3,4))),(NULL), (MAP(LIST_VALUE(1,2),LIST_VALUE(3,4))), (NULL)) as a",
        )



import duckdb
import pytest
from conftest import NumpyPandas, ArrowPandas
from duckdb import Value


class TestDuckDBQuery(object):
    def test_duckdb_query(self, duckdb_cursor):
        # we can use duckdb_cursor.sql to run both DDL statements and select statements
        duckdb_cursor.sql('create view v1 as select 42 i')
        rel = duckdb_cursor.sql('select * from v1')
        assert rel.fetchall()[0][0] == 42

        # also multiple statements
        duckdb_cursor.sql('create view v2 as select i*2 j from v1; create view v3 as select j * 2 from v2;')
        rel = duckdb_cursor.sql('select * from v3')
        assert rel.fetchall()[0][0] == 168

        # we can run multiple select statements - we get only the last result
        res = duckdb_cursor.sql('select 42; select 84;').fetchall()
        assert res == [(84,)]

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_duckdb_from_query_multiple_statements(self, pandas):
        tst_df = pandas.DataFrame({'a': [1, 23, 3, 5]})

        res = duckdb.sql(
            '''
        select 42; select *
        from tst_df
        union all
        select *
        from tst_df;
        '''
        ).fetchall()
        assert res == [(1,), (23,), (3,), (5,), (1,), (23,), (3,), (5,)]

    def test_duckdb_query_empty_result(self):
        con = duckdb.connect()
        # show tables on empty connection does not produce any tuples
        res = con.query('show tables').fetchall()
        assert res == []

    def test_named_param(self):
        con = duckdb.connect()

        original_res = con.execute(
            """
            select
                count(*) FILTER (WHERE i >= $1),
                sum(i) FILTER (WHERE i < $2),
                avg(i) FILTER (WHERE i < $1)
            from
                range(100) tbl(i)
        """,
            [5, 10],
        ).fetchall()

        res = con.execute(
            """
            select
                count(*) FILTER (WHERE i >= $param),
                sum(i) FILTER (WHERE i < $other_param),
                avg(i) FILTER (WHERE i < $param)
            from
                range(100) tbl(i)
        """,
            {'param': 5, 'other_param': 10},
        ).fetchall()

        assert res == original_res

    def test_named_param_not_dict(self):
        con = duckdb.connect()

        with pytest.raises(
            duckdb.InvalidInputException,
            match="Values were not provided for the following prepared statement parameters: name1, name2, name3",
        ):
            con.execute("select $name1, $name2, $name3", ['name1', 'name2', 'name3'])

    def test_named_param_basic(self):
        con = duckdb.connect()

        res = con.execute("select $name1, $name2, $name3", {'name1': 5, 'name2': 3, 'name3': 'a'}).fetchall()
        assert res == [
            (5, 3, 'a'),
        ]

    def test_named_param_not_exhaustive(self):
        con = duckdb.connect()

        with pytest.raises(
            duckdb.InvalidInputException,
            match="Invalid Input Error: Values were not provided for the following prepared statement parameters: name3",
        ):
            con.execute("select $name1, $name2, $name3", {'name1': 5, 'name2': 3})

    def test_named_param_excessive(self):
        con = duckdb.connect()

        with pytest.raises(
            duckdb.InvalidInputException,
            match="Values were not provided for the following prepared statement parameters: name3",
        ):
            con.execute("select $name1, $name2, $name3", {'name1': 5, 'name2': 3, 'not_a_named_param': 5})

    def test_named_param_not_named(self):
        con = duckdb.connect()

        with pytest.raises(
            duckdb.InvalidInputException,
            match="Values were not provided for the following prepared statement parameters: 1, 2",
        ):
            con.execute("select $1, $1, $2", {'name1': 5, 'name2': 3})

    def test_named_param_mixed(self):
        con = duckdb.connect()

        with pytest.raises(
            duckdb.NotImplementedException, match="Mixing named and positional parameters is not supported yet"
        ):
            con.execute("select $name1, $1, $2", {'name1': 5, 'name2': 3})

    def test_named_param_strings_with_dollarsign(self):
        con = duckdb.connect()

        res = con.execute("select '$name1', $name1, $name1, '$name1'", {'name1': 5}).fetchall()
        assert res == [('$name1', 5, 5, '$name1')]

    def test_named_param_case_insensivity(self):
        con = duckdb.connect()

        res = con.execute(
            """
                select $NaMe1, $NAME2, $name3
            """,
            {'name1': 5, 'nAmE2': 3, 'NAME3': 'a'},
        ).fetchall()
        assert res == [
            (5, 3, 'a'),
        ]

    def test_named_param_keyword(self):
        con = duckdb.connect()

        result = con.execute("SELECT $val", {"val": 42}).fetchone()
        assert result == (42,)

        result = con.execute("SELECT $value", {"value": 42}).fetchone()
        assert result == (42,)

    def test_conversion_from_tuple(self):
        con = duckdb.connect()

        # Tuple converts to list
        result = con.execute("select $1", [(21, 22, 42)]).fetchall()
        assert result == [([21, 22, 42],)]

        # If wrapped in a Value, it can convert to a struct
        result = con.execute("select $1", [Value(('a', 21, True), {'a': str, 'b': int, 'c': bool})]).fetchall()
        assert result == [({'a': 'a', 'b': 21, 'c': True},)]

        # If the amount of items in the tuple and the children of the struct don't match
        # we throw an error
        with pytest.raises(
            duckdb.InvalidInputException,
            match='Tried to create a STRUCT value from a tuple containing 3 elements, but the STRUCT consists of 2 children',
        ):
            result = con.execute("select $1", [Value(('a', 21, True), {'a': str, 'b': int})]).fetchall()

        # If we try to create anything other than a STRUCT or a LIST out of the tuple, we throw an error
        with pytest.raises(duckdb.InvalidInputException, match="Can't convert tuple to a Value of type VARCHAR"):
            result = con.execute("select $1", [Value((21, 42), str)])

    def test_column_name_behavior(self, duckdb_cursor):
        _ = pytest.importorskip("pandas")

        expected_names = ['one', 'ONE_1']

        df = duckdb_cursor.execute('select 1 as one, 2 as "ONE"').fetchdf()
        assert expected_names == list(df.columns)

        duckdb_cursor.register('tbl', df)
        df = duckdb_cursor.execute("select * from tbl").fetchdf()
        assert expected_names == list(df.columns)

        df = duckdb_cursor.execute('with t as (select 1 as one, 2 as "ONE") select * from t').fetchdf()
        assert expected_names == list(df.columns)



import pytest
import duckdb


class TestWithPropagatingExceptions(object):
    def test_with(self):
        # Should propagate exception raised in the 'with duckdb.connect() ..'
        with pytest.raises(duckdb.ParserException, match="syntax error at or near *"):
            with duckdb.connect() as con:
                print('before')
                con.execute('invalid')
                print('after')

        # Does not raise an exception
        with duckdb.connect() as con:
            print('before')
            con.execute('select 1')
            print('after')



import duckdb
import os
import datetime
import pytest

try:
    import pyarrow as pa
    import pandas as pd

    can_run = True
except:
    can_run = False


class TestArrowInterval(object):
    def test_duration_types(self, duckdb_cursor):
        if not can_run:
            return
        expected_arrow = pa.Table.from_arrays(
            [pa.array([pa.MonthDayNano([0, 0, 1000000000])], type=pa.month_day_nano_interval())], ['a']
        )
        data = (
            pa.array([1000000000], type=pa.duration('ns')),
            pa.array([1000000], type=pa.duration('us')),
            pa.array([1000], pa.duration('ms')),
            pa.array([1], pa.duration('s')),
        )
        arrow_table = pa.Table.from_arrays([data[0], data[1], data[2], data[3]], ['a', 'b', 'c', 'd'])
        rel = duckdb.from_arrow(arrow_table).arrow()
        assert rel['a'] == expected_arrow['a']
        assert rel['b'] == expected_arrow['a']
        assert rel['c'] == expected_arrow['a']
        assert rel['d'] == expected_arrow['a']

    def test_duration_null(self, duckdb_cursor):
        if not can_run:
            return
        expected_arrow = pa.Table.from_arrays([pa.array([None], type=pa.month_day_nano_interval())], ['a'])
        data = (
            pa.array([None], type=pa.duration('ns')),
            pa.array([None], type=pa.duration('us')),
            pa.array([None], pa.duration('ms')),
            pa.array([None], pa.duration('s')),
        )
        arrow_table = pa.Table.from_arrays([data[0], data[1], data[2], data[3]], ['a', 'b', 'c', 'd'])
        rel = duckdb.from_arrow(arrow_table).arrow()
        assert rel['a'] == expected_arrow['a']
        assert rel['b'] == expected_arrow['a']
        assert rel['c'] == expected_arrow['a']
        assert rel['d'] == expected_arrow['a']

    def test_duration_overflow(self, duckdb_cursor):
        if not can_run:
            return

        # Only seconds can overflow
        data = pa.array([9223372036854775807], pa.duration('s'))
        arrow_table = pa.Table.from_arrays([data], ['a'])

        with pytest.raises(duckdb.ConversionException, match='Could not convert Interval to Microsecond'):
            arrow_from_duck = duckdb.from_arrow(arrow_table).arrow()



import duckdb

try:
    import pyarrow
    import pyarrow.parquet
    import numpy as np

    can_run = True
except:
    can_run = False


def munge(cell):
    try:
        cell = round(float(cell), 2)
    except (ValueError, TypeError):
        cell = str(cell)
    return cell


def check_result(result, answers):
    for q_res in answers:
        db_result = result.fetchone()
        cq_results = q_res.split("|")
        # The end of the rows, continue
        if cq_results == [''] and str(db_result) == 'None' or str(db_result[0]) == 'None':
            continue
        ans_result = [munge(cell) for cell in cq_results]
        db_result = [munge(cell) for cell in db_result]

        assert ans_result == db_result
    return True


class TestTPCHArrow(object):
    def test_tpch_arrow(self, duckdb_cursor):
        if not can_run:
            return

        tpch_tables = ['part', 'partsupp', 'supplier', 'customer', 'lineitem', 'orders', 'nation', 'region']
        arrow_tables = []

        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("CALL dbgen(sf=0.01);")

        for tpch_table in tpch_tables:
            duck_tbl = duckdb_conn.table(tpch_table)
            arrow_tables.append(duck_tbl.arrow())
            duck_arrow_table = duckdb_conn.from_arrow(arrow_tables[-1])
            duckdb_conn.execute("DROP TABLE " + tpch_table)
            duck_arrow_table.create(tpch_table)

        for i in range(1, 23):
            query = duckdb_conn.execute("select query from tpch_queries() where query_nr=" + str(i)).fetchone()[0]
            answers = (
                duckdb_conn.execute(
                    "select answer from tpch_answers() where scale_factor = 0.01 and query_nr=" + str(i)
                )
                .fetchone()[0]
                .split("\n")[1:]
            )
            result = duckdb_conn.execute(query)
            assert check_result(result, answers)
            print("Query " + str(i) + " works")

    def test_tpch_arrow_01(self, duckdb_cursor):
        if not can_run:
            return

        tpch_tables = ['part', 'partsupp', 'supplier', 'customer', 'lineitem', 'orders', 'nation', 'region']
        arrow_tables = []

        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("CALL dbgen(sf=0.1);")

        for tpch_table in tpch_tables:
            duck_tbl = duckdb_conn.table(tpch_table)
            arrow_tables.append(duck_tbl.arrow())
            duck_arrow_table = duckdb_conn.from_arrow(arrow_tables[-1])
            duckdb_conn.execute("DROP TABLE " + tpch_table)
            duck_arrow_table.create(tpch_table)

        for i in range(1, 23):
            query = duckdb_conn.execute("select query from tpch_queries() where query_nr=" + str(i)).fetchone()[0]
            answers = (
                duckdb_conn.execute("select answer from tpch_answers() where scale_factor = 0.1 and query_nr=" + str(i))
                .fetchone()[0]
                .split("\n")[1:]
            )
            result = duckdb_conn.execute(query)
            assert check_result(result, answers)
            print("Query " + str(i) + " works")

    def test_tpch_arrow_batch(self, duckdb_cursor):
        if not can_run:
            return

        tpch_tables = ['part', 'partsupp', 'supplier', 'customer', 'lineitem', 'orders', 'nation', 'region']
        arrow_tables = []

        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("CALL dbgen(sf=0.01);")

        for tpch_table in tpch_tables:
            duck_tbl = duckdb_conn.table(tpch_table)
            arrow_tables.append(pyarrow.Table.from_batches(duck_tbl.arrow().to_batches(10)))
            duck_arrow_table = duckdb_conn.from_arrow(arrow_tables[-1])
            duckdb_conn.execute("DROP TABLE " + tpch_table)
            duck_arrow_table.create(tpch_table)

        for i in range(1, 23):
            query = duckdb_conn.execute("select query from tpch_queries() where query_nr=" + str(i)).fetchone()[0]
            answers = (
                duckdb_conn.execute(
                    "select answer from tpch_answers() where scale_factor = 0.01 and query_nr=" + str(i)
                )
                .fetchone()[0]
                .split("\n")[1:]
            )
            result = duckdb_conn.execute(query)
            assert check_result(result, answers)
            print("Query " + str(i) + " works")

        duckdb_conn.execute("PRAGMA threads=4")
        duckdb_conn.execute("PRAGMA verify_parallelism")

        for i in range(1, 23):
            query = duckdb_conn.execute("select query from tpch_queries() where query_nr=" + str(i)).fetchone()[0]
            answers = (
                duckdb_conn.execute(
                    "select answer from tpch_answers() where scale_factor = 0.01 and query_nr=" + str(i)
                )
                .fetchone()[0]
                .split("\n")[1:]
            )
            result = duckdb_conn.execute(query)
            assert check_result(result, answers)
            print("Query " + str(i) + " works (Parallel)")



import duckdb
import pytest
import os
import pandas as pd

try:
    import pyarrow.parquet as pq
    import pyarrow.dataset as ds

    can_run = True
except:
    can_run = False


class TestArrowReplacementScan(object):
    def test_arrow_table_replacement_scan(self, duckdb_cursor):
        if not can_run:
            return

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')
        userdata_parquet_table = pq.read_table(parquet_filename)
        df = userdata_parquet_table.to_pandas()

        con = duckdb.connect()

        for i in range(5):
            assert con.execute("select count(*) from userdata_parquet_table").fetchone() == (1000,)
            assert con.execute("select count(*) from df").fetchone() == (1000,)

    def test_arrow_table_replacement_scan_view(self, duckdb_cursor):
        if not can_run:
            return

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')
        userdata_parquet_table = pq.read_table(parquet_filename)

        con = duckdb.connect()

        con.execute("create view x as select * from userdata_parquet_table")
        del userdata_parquet_table
        with pytest.raises(duckdb.CatalogException, match='Table with name userdata_parquet_table does not exist'):
            assert con.execute("select count(*) from x").fetchone()

    def test_arrow_dataset_replacement_scan(self, duckdb_cursor):
        if not can_run:
            return
        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')
        userdata_parquet_table = pq.read_table(parquet_filename)
        userdata_parquet_dataset = ds.dataset(parquet_filename)

        con = duckdb.connect()
        assert con.execute("select count(*) from userdata_parquet_dataset").fetchone() == (1000,)



import os

script_dir = os.path.dirname(__file__)
from typing import List, Dict
import json

# Load existing JSON data from a file if it exists
json_data = {}
json_cache_path = os.path.join(script_dir, "cache_data.json")
try:
    with open(json_cache_path, "r") as file:
        json_data = json.load(file)
except FileNotFoundError:
    print("Please first use 'generate_import_cache_json.py' first to generate json")


# deal with leaf nodes?? Those are just PythonImportCacheItem
def get_class_name(path: str) -> str:
    parts: List[str] = path.replace('_', '').split('.')
    parts = [x.title() for x in parts]
    return ''.join(parts) + 'CacheItem'


def get_filename(name: str) -> str:
    return name.replace('_', '').lower() + '_module.hpp'


def get_variable_name(name: str) -> str:
    if name in ['short', 'ushort']:
        return name + '_'
    return name


def collect_items_of_module(module: dict, collection: Dict):
    global json_data
    children = module['children']
    collection[module['full_path']] = module
    for child in children:
        collect_items_of_module(json_data[child], collection)


class CacheItem:
    def __init__(self, module: dict, items):
        self.name = module['name']
        self.module = module
        self.items = items
        self.class_name = get_class_name(module['full_path'])

    def get_full_module_path(self):
        if self.module['type'] != 'module':
            return ''
        full_path = self.module['full_path']
        return f"""
public:
\tstatic constexpr const char *Name = "{full_path}";
"""

    def get_optionally_required(self):
        if 'required' not in self.module:
            return ''
        string = f"""
protected:
\tbool IsRequired() const override final {{
\t\treturn {str(self.module['required']).lower()};
\t}}
"""
        return string

    def get_variables(self):
        variables = []
        for key in self.module['children']:
            item = self.items[key]
            name = item['name']
            var_name = get_variable_name(name)
            if item['children'] == []:
                class_name = 'PythonImportCacheItem'
            else:
                class_name = get_class_name(item['full_path'])
            variables.append(f'\t{class_name} {var_name};')
        return '\n'.join(variables)

    def get_initializer(self):
        variables = []
        for key in self.module['children']:
            item = self.items[key]
            name = item['name']
            var_name = get_variable_name(name)
            if item['children'] == []:
                initialization = f'{var_name}("{name}", this)'
                variables.append(initialization)
            else:
                if item['type'] == 'module':
                    arguments = ''
                else:
                    arguments = 'this'
                initialization = f'{var_name}({arguments})'
                variables.append(initialization)
        if self.module['type'] != 'module':
            constructor_params = f'"{self.name}"'
            constructor_params += ', parent'
        else:
            full_path = self.module['full_path']
            constructor_params = f'"{full_path}"'
        return f'PythonImportCacheItem({constructor_params}), ' + ', '.join(variables) + '{}'

    def get_constructor(self):
        if self.module['type'] == 'module':
            return f'{self.class_name}()'
        return f'{self.class_name}(optional_ptr<PythonImportCacheItem> parent)'

    def to_string(self):
        return f"""
struct {self.class_name} : public PythonImportCacheItem {{
{self.get_full_module_path()}
public:
\t{self.get_constructor()} : {self.get_initializer()}
\t~{self.class_name}() override {{}}

{self.get_variables()}
{self.get_optionally_required()}
}};
"""


def collect_classes(items: Dict) -> List:
    output: List = []
    for item in items.values():
        if item['children'] == []:
            continue
        output.append(CacheItem(item, items))
    return output


class ModuleFile:
    def __init__(self, module: dict):
        self.module = module
        self.file_name = get_filename(module['name'])
        self.items = {}
        collect_items_of_module(module, self.items)
        self.classes = collect_classes(self.items)
        self.classes.reverse()

    def get_classes(self):
        classes = []
        for item in self.classes:
            classes.append(item.to_string())
        return ''.join(classes)

    def to_string(self):
        string = f"""
//===----------------------------------------------------------------------===//
//                         DuckDB
//
// duckdb_python/import_cache/modules/{self.file_name}
//
//
//===----------------------------------------------------------------------===//

#pragma once

#include "duckdb_python/import_cache/python_import_cache_item.hpp"

namespace duckdb {{
{self.get_classes()}
}} // namespace duckdb
"""
        return string


files: List[ModuleFile] = []
for name, value in json_data.items():
    if value['full_path'] != value['name']:
        continue
    files.append(ModuleFile(value))

for file in files:
    content = file.to_string()
    path = f'src/include/duckdb_python/import_cache/modules/{file.file_name}'
    import_cache_path = os.path.join(script_dir, '..', path)
    with open(import_cache_path, "w") as f:
        f.write(content)


def get_root_modules(files: List[ModuleFile]):
    modules = []
    for file in files:
        name = file.module['name']
        class_name = get_class_name(name)
        modules.append(f'\t{class_name} {name};')
    return '\n'.join(modules)


# Generate the python_import_cache.hpp file
# adding all the root modules with their 'name'
import_cache_file = f"""
//===----------------------------------------------------------------------===//
//                         DuckDB
//
// duckdb_python/import_cache/python_import_cache.hpp
//
//
//===----------------------------------------------------------------------===//

#pragma once

#include "duckdb_python/pybind11/pybind_wrapper.hpp"
#include "duckdb.hpp"
#include "duckdb/common/vector.hpp"
#include "duckdb_python/import_cache/python_import_cache_modules.hpp"

namespace duckdb {{

struct PythonImportCache {{
public:
    explicit PythonImportCache() {{
    }}
    ~PythonImportCache();

public:
{get_root_modules(files)}

public:
    py::handle AddCache(py::object item);

private:
    vector<py::object> owned_objects;
}};

}} // namespace duckdb

"""

import_cache_path = os.path.join(script_dir, '..', 'src/include/duckdb_python/import_cache/python_import_cache.hpp')
with open(import_cache_path, "w") as f:
    f.write(import_cache_file)


def get_module_file_path_includes(files: List[ModuleFile]):
    includes = []
    for file in files:
        includes.append(f'#include "duckdb_python/import_cache/modules/{file.file_name}"')
    return '\n'.join(includes)


module_includes = get_module_file_path_includes(files)

modules_header = os.path.join(
    script_dir, '..', 'src/include/duckdb_python/import_cache/python_import_cache_modules.hpp'
)
with open(modules_header, "w") as f:
    f.write(module_includes)

# Generate the python_import_cache_modules.hpp file
# listing all the generated header files
